import pygame

from pygame_gui.core.ui_element import UIElement


class UIContainer(UIElement):
    """
    A UI Container holds any number of other UI elements inside of a rectangle. When we move the UIContainer
    all the UI elements contained within it can be moved as well.

    This class helps us make UI Windows, but likely will have wider uses as well as the GUI system develops.

    :param relative_rect: A pygame.Rect whose position is relative to whatever UIContainer it is inside of, if any.
    :param manager: The UIManager that manages this UIElement.
    :param container: The UIContainer that this UIElement is contained within.
    :param parent_element: The element this element 'belongs to' in the theming hierarchy.
    :param object_id: A custom defined ID for fine tuning of theming.
    """
    def __init__(self, relative_rect, manager,
                 container=None, parent_element=None, object_id=None):

        self._layer = 0
        self.ui_manager = manager

        new_element_ids, new_object_ids = self.create_valid_ids(parent_element=parent_element,
                                                                object_id=object_id,
                                                                element_id='container')

        super().__init__(relative_rect, manager, container,
                         object_ids=new_object_ids,
                         element_ids=new_element_ids,
                         starting_height=1,
                         layer_thickness=1)

        self.sprite_group = self.ui_manager.get_sprite_group()
        self.image = pygame.Surface((0, 0))
        self.elements = []
        self.layer_thickness = 1

        self.hovered = False

    def add_element(self, element):
        """
        Add a UIElement to the container. The UI's relative_rect parameter will be relative to this container.

        :param element: A UIElement to add to this container.
        """
        element.change_layer(self._layer + element.starting_height)
        self.elements.append(element)
        self.recalculate_container_layer_thickness()

    def remove_element(self, element):
        """
        Remove a UIElement from this container.

        :param element: A UIElement to remove from this container.
        """
        if element in self.elements:
            self.elements.remove(element)
        self.recalculate_container_layer_thickness()

    def recalculate_container_layer_thickness(self):
        """
        This function will iterate through the elements in our container and determine the maximum 'height'
        that they reach in the 'layer stack'. We then use that to determine the overall 'thickness' of this container.
        The thickness value is used to determine where to place overlapping windows in the layers
        """
        max_element_top_layer = 0
        for element in self.elements:
            if element.top_layer > max_element_top_layer:
                max_element_top_layer = element.top_layer

        self.layer_thickness = max_element_top_layer - self._layer

    def change_container_layer(self, new_layer):
        """
        Change the layer of this container. Layers are used by the GUI to control the order in which things are drawn
        and which things should currently be interactive (so you can't interact with things behind other things).

        This particular method is most often used to shift the visible contents of a window in front of any others when
        it is moved to the front of the window stack.

        :param new_layer: The layer to move our container to.
        """
        if new_layer != self._layer:
            self._layer = new_layer
            self.sprite_group.change_layer(self, self._layer)
            for element in self.elements:
                element.change_layer(self._layer + element.starting_height)

    def update_containing_rect_position(self):
        """
        This function is called when we move the container to update all the contained UI Elements to move as well.
        """
        super().update_containing_rect_position()
        for element in self.elements:
            element.update_containing_rect_position()

    def get_top_layer(self):
        """
        Assuming we have correctly calculated the 'thickness' of this container, this method will return the 'highest'
        layer in the LayeredUpdates UI Group.

        :return int: An integer representing the current highest layer being used by this container.
        """
        return self._layer + self.layer_thickness

    def kill(self):
        """
        Overrides the standard kill method of UI Elements (and pygame sprites beyond that) to also call the kill method
        on all contained UI Elements.
        """
        self.clear()
        super().kill()

    def clear(self):
        """
        Removes and kills all the UI elements inside this container.
        """
        while len(self.elements) > 0:
            self.elements.pop().kill()

    # noinspection PyUnusedLocal
    def check_hover(self, time_delta, hovered_higher_element):
        """
        A method that helps us to determine which, if any, UI Element is currently being hovered by the mouse.

        :param time_delta: A float, the time in seconds between the last call to this function and now (roughly).
        :param hovered_higher_element: A boolean, representing whether we have already hovered a 'higher' element.
        :return bool: A boolean that is true if we have hovered a UI element, either just now or before this method.
        """
        if self.alive():
            mouse_x, mouse_y = pygame.mouse.get_pos()

            if self.rect.collidepoint(mouse_x, mouse_y) and not hovered_higher_element:
                if not self.hovered:
                    self.hovered = True
                hovered_higher_element = True

            else:
                if self.hovered:
                    self.hovered = False

        elif self.hovered:
            self.hovered = False
        return hovered_higher_element



import platform
import subprocess
import time
import contextlib

# noinspection SpellCheckingInspection
"""
This code owes a lot to pyperclip by Al Sweigart al@inventwithpython.com.

I would just use the damn library but working in schools frequently requires
minimal dependencies.
"""

plat = platform.system().upper()
if plat == 'WINDOWS':
    import ctypes
    from ctypes import c_size_t, sizeof, c_wchar_p, c_wchar
    from ctypes.wintypes import HGLOBAL, LPVOID, BOOL, UINT, HANDLE, HWND, DWORD, INT, HMENU, HINSTANCE, LPCSTR


    @contextlib.contextmanager
    def __windows_clipboard(hwnd):
        ctypes.windll.user32.OpenClipboard.argtypes = [HWND]
        ctypes.windll.user32.OpenClipboard.restype = BOOL

        ctypes.windll.user32.CloseClipboard.argtypes = []
        ctypes.windll.user32.CloseClipboard.restype = BOOL

        t = time.time() + 0.5
        success = False
        while time.time() < t:
            success = ctypes.windll.user32.OpenClipboard(hwnd)
            if success:
                break
            time.sleep(0.01)
        if not success:
            raise Exception

        try:
            yield
        finally:
            ctypes.windll.user32.CloseClipboard()


    def __windows_paste():
        ctypes.windll.user32.GetClipboardData.argtypes = [UINT]
        ctypes.windll.user32.GetClipboardData.restype = HANDLE
        cf_unicode_text = 13
        with __windows_clipboard(None):
            handle = ctypes.windll.user32.GetClipboardData(cf_unicode_text)
            if not handle:
                return ""
            return ctypes.c_wchar_p(handle).value


    # noinspection PyUnresolvedReferences
    class CheckedCall(object):

        def __init__(self, f):
            super(CheckedCall, self).__setattr__("f", f)

        def __call__(self, *args):
            ret = self.f(*args)

            return ret

        def __setattr__(self, key, value):
            setattr(self.f, key, value)


    def __windows_copy(data: str):
        msvcrt = ctypes.CDLL('msvcrt')

        safe_create_window = CheckedCall(ctypes.windll.user32.CreateWindowExA)
        safe_create_window.argtypes = [DWORD, LPCSTR, LPCSTR, DWORD, INT, INT,
                                       INT, INT, HWND, HMENU, HINSTANCE, LPVOID]
        safe_create_window.restype = HWND

        safe_destroy_window = CheckedCall(ctypes.windll.user32.DestroyWindow)
        safe_destroy_window.argtypes = [HWND]
        safe_destroy_window.restype = BOOL

        safe_empty = ctypes.windll.user32.EmptyClipboard
        safe_empty.argtypes = []
        safe_empty.restype = BOOL

        safe_alloc = CheckedCall(ctypes.windll.kernel32.GlobalAlloc)
        safe_alloc.argtypes = [UINT, c_size_t]
        safe_alloc.restype = HGLOBAL

        safe_lock = CheckedCall(ctypes.windll.kernel32.GlobalLock)
        safe_lock.argtypes = [HGLOBAL]
        safe_lock.restype = LPVOID

        safe_unlock = CheckedCall(ctypes.windll.kernel32.GlobalUnlock)
        safe_unlock.argtypes = [HGLOBAL]
        safe_unlock.restype = BOOL

        safe_set_clipboard = CheckedCall(ctypes.windll.user32.SetClipboardData)
        safe_set_clipboard.argtypes = [UINT, HANDLE]
        safe_set_clipboard.restype = HANDLE

        wcslen = msvcrt.wcslen
        wcslen.argtypes = [c_wchar_p]
        wcslen.restype = UINT

        gmem_moveable = 0x0002
        cf_unicode_text = 13

        # weirdly this temporary window handle seems to work for pasting where the
        # normal pygame window handle does not
        hwnd = safe_create_window(0, b"STATIC", None, 0, 0, 0, 0, 0,
                                  None, None, None, None)

        data = str(data)
        with __windows_clipboard(hwnd):
            safe_empty()

            if data:
                count = wcslen(data) + 1
                handle = safe_alloc(gmem_moveable, count * sizeof(c_wchar))
                locked_handle = safe_lock(handle)

                ctypes.memmove(c_wchar_p(locked_handle), c_wchar_p(data), count * sizeof(c_wchar))

                safe_unlock(handle)
                safe_set_clipboard(cf_unicode_text, handle)

        safe_destroy_window(hwnd)
elif plat == 'LINUX':

    def __linux_copy(data: str):
        process = subprocess.Popen(['xsel', '-b', '-i'], stdin=subprocess.PIPE, close_fds=True)
        process.communicate(input=data.encode('utf-8'))


    def __linux_paste():
        process = subprocess.Popen(['xsel', '-b', '-o'], stdout=subprocess.PIPE, close_fds=True)
        stdout, stderr = process.communicate()
        return stdout.decode('utf-8')

else:
    def __mac_copy(data: str):
        process = subprocess.Popen(
            'pbcopy', env={'LANG': 'en_US.UTF-8'}, stdin=subprocess.PIPE)
        process.communicate(data.encode('utf-8'))


    def __mac_paste():
        return subprocess.check_output(
            'pbpaste', env={'LANG': 'en_US.UTF-8'}).decode('utf-8')


def clipboard_copy(data: str):
    current_platform = platform.system().upper()
    if current_platform == 'WINDOWS':
        __windows_copy(data)
    elif current_platform == 'LINUX':
        __linux_copy(data)
    else:
        __mac_copy(data)


def clipboard_paste():
    current_platform = platform.system().upper()
    if current_platform == 'WINDOWS':
        return __windows_paste()
    elif current_platform == 'LINUX':
        return __linux_paste()
    else:
        return __mac_paste()



import pygame
import re
import warnings
from typing import Union, List

import pygame_gui
from pygame_gui import ui_manager
from pygame_gui.core import ui_container
from pygame_gui.core.utility import clipboard_paste, clipboard_copy
from pygame_gui.core.ui_element import UIElement
from pygame_gui.core.drawable_shapes import RectDrawableShape, RoundedRectangleShape
from pygame_gui.core.ui_appearance_theme import ColourGradient


class UITextEntryLine(UIElement):
    """
    A GUI element for text entry from a keyboard, on a single line. The element supports the standard
    copy and paste keyboard shortcuts CTRL+V, CTRL+C & CTRL+X as well as CTRL+A.

    There are methods that allow the entry element to restrict the characters that can be input into the text box

    The height of the text entry line element will be determined by the font used rather than the standard method
    for UIElements of just using the height of the input rectangle.

    :param relative_rect: A rectangle describing the position and width of the text entry element.
    :param manager: The UIManager that manages this element.
    :param container: The container that this element is within. If set to None will be the root window's container.
    :param parent_element: The element this element 'belongs to' in the theming hierarchy.
    :param object_id: A custom defined ID for fine tuning of theming.
    """

    _number_character_set = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

    # excluding these characters won't ensure that user entered text is a valid filename but they can help
    # reduce the problems that input will leave you with.
    _forbidden_file_path_characters = ['<', '>', ':', '"', '/', '\\', '|', '?', '*', '\0', '.']

    def __init__(self, relative_rect: pygame.Rect,
                 manager: ui_manager.UIManager,
                 container: ui_container.UIContainer = None,
                 parent_element: UIElement = None,
                 object_id: Union[str, None] = None):

        new_element_ids, new_object_ids = self.create_valid_ids(parent_element=parent_element,
                                                                object_id=object_id,
                                                                element_id='text_entry_line')
        super().__init__(relative_rect, manager, container,
                         starting_height=1, layer_thickness=1,
                         element_ids=new_element_ids,
                         object_ids=new_object_ids)
        self.selected = False

        self.text = ""

        # theme font
        self.font = None

        self.shadow_width = None
        self.border_width = None
        self.horiz_line_padding = None
        self.vert_line_padding = None
        self.text_surface = None
        self.cursor = None
        self.background_and_border = None
        self.text_image_rect = None
        self.text_image = None

        # colours from theme
        self.background_colour = None
        self.text_colour = None
        self.selected_text_colour = None
        self.selected_bg_colour = None
        self.border_colour = None
        self.padding = None

        self.drawable_shape = None
        self.shape_type = 'rectangle'
        self.shape_corner_radius = None

        # input timings - I expect nobody really wants to mess with these that much
        # ideally we could populate from the os settings but that sounds like a headache
        self.key_repeat = 0.5
        self.cursor_blink_delay_after_moving_acc = 0.0
        self.cursor_blink_delay_after_moving = 1.0
        self.blink_cursor_time_acc = 0.0
        self.blink_cursor_time = 0.4
        self.double_click_select_time = 0.4
        self.double_click_select_time_acc = 0.0

        self.start_text_offset = 0
        self.edit_position = 0
        self.select_range = [0, 0]
        self.selection_in_progress = False

        self.cursor_on = False
        self.cursor_has_moved_recently = False
        self.double_click_started = False

        # restrictions on text input
        self.allowed_characters = None
        self.forbidden_characters = None
        self.length_limit = None

        self.rebuild_from_changed_theme_data()

    def rebuild(self):
        """
        Rebuild whatever needs building.

        """
        line_height = self.font.size(' ')[1]

        self.relative_rect.height = (line_height + (2 * self.vert_line_padding) +
                                     (2 * self.border_width) + (2 * self.shadow_width))
        self.rect.height = self.relative_rect.height

        self.text_image_rect = pygame.Rect((self.border_width + self.shadow_width + self.shape_corner_radius,
                                            self.border_width + self.shadow_width),
                                           (self.rect.width - (self.border_width * 2) - (self.shadow_width * 2) -
                                            (2 * self.shape_corner_radius),
                                            self.rect.height - (self.border_width * 2) - (self.shadow_width * 2)))

        theming_parameters = {'normal_bg': self.background_colour,
                              'normal_border': self.border_colour,
                              'border_width': self.border_width,
                              'shadow_width': self.shadow_width,
                              'shape_corner_radius': self.shape_corner_radius}

        if self.shape_type == 'rectangle':
            self.drawable_shape = RectDrawableShape(self.rect, theming_parameters,
                                                    ['normal'], self.ui_manager)
        elif self.shape_type == 'rounded_rectangle':
            self.drawable_shape = RoundedRectangleShape(self.rect, theming_parameters,
                                                        ['normal'], self.ui_manager)

        self.background_and_border = self.drawable_shape.get_surface('normal')

        self.text_image = pygame.Surface(self.text_image_rect.size, flags=pygame.SRCALPHA, depth=32)
        if type(self.background_colour) == ColourGradient:
            self.text_image.fill(pygame.Color("#FFFFFFFF"))
            self.background_colour.apply_gradient_to_surface(self.text_image)
        else:
            self.text_image.fill(self.background_colour)

        self.image = self.background_and_border.copy()

        self.cursor = pygame.Rect((self.text_image_rect.x +
                                   self.horiz_line_padding - self.start_text_offset,
                                   self.text_image_rect.y +
                                   self.vert_line_padding), (1, line_height))

        # setup for drawing
        self.redraw()

    def set_text_length_limit(self, limit: int):
        """
        Allows a character limit to be set on the text entry element. By default there is no limit on the number
        of characters that can be entered.

        :param limit: The character limit as an integer.
        """
        self.length_limit = limit

    def get_text(self) -> str:
        """
        Gets the text in the entry line element.

        :return: A string .
        """
        return self.text

    def set_text(self, text: str):
        """
        Allows the text displayed in the text entry element to be set via code. Useful for setting an initial or
        existing value that is able to be edited.

        The string to set must be valid for the text entry element for this to work.

        :param text: The text string to set.
        """
        if self.validate_text_string(text):
            within_length_limit = True
            if self.length_limit is not None:
                if len(text) > self.length_limit:
                    within_length_limit = False
            if within_length_limit:
                self.text = text
                self.edit_position = len(self.text)
                self.cursor_has_moved_recently = True
            else:
                warnings.warn("Tried to set text string that is too long on text entry element")
        else:
            warnings.warn("Tried to set text string with invalid characters on text entry element")

    def redraw(self):
        """
        Redraws the entire text entry element onto the underlying sprite image. Usually called when the displayed text
        has been edited or changed in some fashion.
        """
        if type(self.background_colour) == ColourGradient:
            self.text_image.fill(pygame.Color("#FFFFFFFF"))
            self.background_colour.apply_gradient_to_surface(self.text_image)
        else:
            self.text_image.fill(self.background_colour)
        if self.select_range[0] != self.select_range[1]:
            low_end = min(self.select_range[0], self.select_range[1])
            high_end = max(self.select_range[0], self.select_range[1])
            pre_select_area_text = self.text[:low_end]
            select_area_text = self.text[low_end:high_end]
            post_select_area_text = self.text[high_end:]

            pre_select_area_surface = None
            post_select_area_surface = None

            if len(pre_select_area_text) > 0:
                if type(self.text_colour) != ColourGradient:
                    pre_select_area_surface = self.font.render(pre_select_area_text, True, self.text_colour)
                else:
                    pre_select_area_surface = self.font.render(pre_select_area_text, True, pygame.Color('#FFFFFFFF'))
                    self.text_colour.apply_gradient_to_surface(pre_select_area_surface)
                width_pre = pre_select_area_surface.get_rect().width
            else:
                width_pre = 0

            if type(self.selected_bg_colour) == ColourGradient:
                select_size = self.font.size(select_area_text)
                select_area_surface = pygame.Surface(select_size, flags=pygame.SRCALPHA, depth=32)
                select_area_surface.fill(pygame.Color('#FFFFFFFF'))
                self.selected_bg_colour.apply_gradient_to_surface(select_area_surface)
                if type(self.selected_text_colour) != ColourGradient:
                    alpha_text = self.font.render(select_area_text, True, self.selected_text_colour)
                else:
                    alpha_text = self.font.render(select_area_text, True, pygame.Color('#FFFFFFFF'))
                    self.selected_text_colour.apply_gradient_to_surface(alpha_text)
                select_area_surface.blit(alpha_text, (0, 0))
            else:
                if type(self.selected_text_colour) != ColourGradient:
                    select_area_surface = self.font.render(select_area_text, True,
                                                           self.selected_text_colour, self.selected_bg_colour)
                else:
                    select_size = self.font.size(select_area_text)
                    select_area_surface = pygame.Surface(select_size, flags=pygame.SRCALPHA, depth=32)
                    select_area_surface.fill(self.selected_bg_colour)

                    alpha_text = self.font.render(select_area_text, True, pygame.Color('#FFFFFFFF'))
                    self.selected_text_colour.apply_gradient_to_surface(alpha_text)
                    select_area_surface.blit(alpha_text, (0, 0))

            if len(post_select_area_text) > 0:
                if type(self.text_colour) != ColourGradient:
                    post_select_area_surface = self.font.render(post_select_area_text, True, self.text_colour)
                else:
                    post_select_area_surface = self.font.render(post_select_area_text, True, pygame.Color('#FFFFFFFF'))
                    self.text_colour.apply_gradient_to_surface(post_select_area_surface)
                width_post = post_select_area_surface.get_rect().width
            else:
                width_post = 0

            text_height = select_area_surface.get_rect().height
            width_select = select_area_surface.get_rect().width

            self.text_surface = pygame.Surface((width_pre + width_select + width_post, text_height),
                                               flags=pygame.SRCALPHA, depth=32)
            if type(self.background_colour) == ColourGradient:
                self.text_image.fill(pygame.Color("#FFFFFFFF"))
                self.background_colour.apply_gradient_to_surface(self.text_image)
            else:
                self.text_image.fill(self.background_colour)
            if pre_select_area_surface is not None:
                self.text_surface.blit(pre_select_area_surface, (0, 0))
            self.text_surface.blit(select_area_surface, (width_pre, 0))
            if post_select_area_surface is not None:
                self.text_surface.blit(post_select_area_surface, (width_pre+width_select, 0))
        else:
            if type(self.text_colour) != ColourGradient:
                self.text_surface = self.font.render(self.text, True, self.text_colour)
            else:
                self.text_surface = self.font.render(self.text, True, pygame.Color('#FFFFFFFF'))
                self.text_colour.apply_gradient_to_surface(self.text_surface)

        text_clip_width = (self.rect.width - (self.horiz_line_padding * 2) - (self.shape_corner_radius * 2) -
                           (self.border_width * 2) - (self.shadow_width * 2))
        text_clip_height = (self.rect.height - (self.vert_line_padding * 2) -
                            (self.border_width * 2) - (self.shadow_width * 2))
        text_surface_clip = pygame.Rect((0,
                                         0),
                                        (text_clip_width,
                                         text_clip_height))

        edit_pos_str = self.text[:self.edit_position]
        width_to_edit_pos = self.font.size(edit_pos_str)[0]

        if self.start_text_offset > width_to_edit_pos:
            total_width_minus_visible = width_to_edit_pos
            self.start_text_offset = total_width_minus_visible
        elif width_to_edit_pos > (self.start_text_offset + text_clip_width):
            total_width_minus_visible = max(0, width_to_edit_pos - text_clip_width)
            self.start_text_offset = total_width_minus_visible
        elif width_to_edit_pos == 0:
            self.start_text_offset = 0
        else:
            pass

        text_surface_clip.x += self.start_text_offset

        if len(self.text) > 0:
            self.text_image.blit(self.text_surface,
                                 (self.horiz_line_padding,
                                  self.vert_line_padding),
                                 text_surface_clip)

        self.redraw_cursor()

    def redraw_cursor(self):
        """
        Redraws only the blinking edit cursor. This allows us to blink the cursor on and off without spending time
        redrawing all the text.
        """
        self.image = self.background_and_border.copy()
        self.image.blit(self.text_image, self.text_image_rect)
        if self.cursor_on:
            cursor_len_str = self.text[:self.edit_position]
            cursor_size = self.font.size(cursor_len_str)
            self.cursor.x = (cursor_size[0] + self.text_image_rect.x +
                             self.horiz_line_padding - self.start_text_offset)

            if type(self.text_colour) != ColourGradient:
                pygame.draw.rect(self.image, self.text_colour, self.cursor)
            else:
                cursor_surface = pygame.Surface(self.cursor.size)
                cursor_surface.fill(pygame.Color('#FFFFFFFF'))
                self.text_colour.apply_gradient_to_surface(cursor_surface)
                self.image.blit(cursor_surface, self.cursor)

    def update(self, time_delta: float):
        """
        Called every update loop of our UI Manager. Largely handles text drag selection and making sure our
        edit cursor blinks on and off.

        :param time_delta: The time in seconds between this update method call and the previous one.
        :return:
        """
        if self.alive():
            if self.double_click_started and self.double_click_select_time_acc < self.double_click_select_time:
                self.double_click_select_time_acc += time_delta
            if self.selection_in_progress:
                mouse_x, mouse_y = pygame.mouse.get_pos()
                select_end_pos = self.find_edit_position_from_pixel_pos(self.start_text_offset + mouse_x)
                new_range = [self.select_range[0], select_end_pos]

                if new_range[0] != self.select_range[0] or new_range[1] != self.select_range[1]:
                    self.select_range[0] = new_range[0]
                    self.select_range[1] = new_range[1]
                    self.edit_position = self.select_range[1]
                    self.cursor_has_moved_recently = True
                        
            if self.cursor_has_moved_recently:
                self.cursor_has_moved_recently = False
                self.cursor_blink_delay_after_moving_acc = 0.0
                self.cursor_on = True
                self.redraw()

            if self.cursor_blink_delay_after_moving_acc > self.cursor_blink_delay_after_moving:
                if self.blink_cursor_time_acc >= self.blink_cursor_time:
                    self.blink_cursor_time_acc = 0.0
                    if self.cursor_on:
                        self.cursor_on = False
                        self.redraw_cursor()
                    elif self.selected:
                        self.cursor_on = True
                        self.redraw_cursor()
                else:
                    self.blink_cursor_time_acc += time_delta
            else:
                self.cursor_blink_delay_after_moving_acc += time_delta

    def unselect(self):
        """
        Called when this element is no longer the current focus.
        """
        self.selected = False
        pygame.key.set_repeat()
        self.select_range = [0, 0]
        self.redraw()

    def select(self):
        """
        Called when we 'select focus' on this element. In this case it sets up the keyboard to repeat held key presses,
        useful for natural feeling keyboard input.
        """
        self.selected = True
        pygame.key.set_repeat(500, 25)
        self.redraw()

    def process_event(self, event: pygame.event.Event) -> bool:
        """
        Allows the text entry box to react to input events, which is it's primary function. The entry element reacts to
        various types of mouse clicks (double click selecting words, drag select), keyboard combos (CTRL+C, CTRL+V,
        CTRL+X, CTRL+A), individual editing keys (Backspace, Delete, Left & Right arrows) and other keys for inputting
        letters, symbols and numbers.

        :param event: The current event to consider reacting to.
        :return bool: Returns True if we've done something with the input event.
        """
        processed_event = False
        if event.type == pygame.MOUSEBUTTONDOWN:
            if event.button == 1:
                mouse_x, mouse_y = event.pos
                if self.rect.collidepoint(mouse_x, mouse_y):
                    if self.double_click_started and self.double_click_select_time_acc < self.double_click_select_time:
                        self.double_click_started = False
                        pattern = re.compile(r"[\w']+")
                        index = min(self.edit_position, len(self.text)-1)
                        if index > 0:
                            char = self.text[index]
                            while not pattern.match(char):
                                index -= 1
                                if index > 0:
                                    char = self.text[index]
                                else:
                                    break
                            while pattern.match(char):
                                index -= 1
                                if index > 0:
                                    char = self.text[index]
                                else:
                                    break
                            if index > 0:
                                start_select_index = index+1
                            else:
                                start_select_index = index
                            index += 1
                            char = self.text[index]
                            while index < len(self.text) and pattern.match(char):
                                index += 1
                                if index < len(self.text):
                                    char = self.text[index]
                            end_select_index = index

                            self.select_range[0] = start_select_index
                            self.select_range[1] = end_select_index
                            self.edit_position = end_select_index
                            self.cursor_has_moved_recently = True
                    else:
                        self.double_click_started = True
                        self.edit_position = self.find_edit_position_from_pixel_pos(self.start_text_offset + mouse_x)
                        self.select_range[0] = self.edit_position
                        self.select_range[1] = self.edit_position
                        self.cursor_has_moved_recently = True
                        self.selection_in_progress = True
                        self.double_click_select_time_acc = 0.0

                    processed_event = True
        if event.type == pygame.MOUSEBUTTONUP:
            if event.button == 1:
                mouse_x, mouse_y = event.pos
                if self.rect.collidepoint(mouse_x, mouse_y):
                    processed_event = True
                    new_edit_pos = self.find_edit_position_from_pixel_pos(self.start_text_offset + mouse_x)
                    if new_edit_pos != self.edit_position:
                        self.edit_position = new_edit_pos
                        self.cursor_has_moved_recently = True
                        self.select_range[1] = self.edit_position
                    self.selection_in_progress = False
                else:
                    self.selection_in_progress = False

        if self.selected:
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_RETURN:
                    entry_finished_event = pygame.event.Event(pygame.USEREVENT,
                                                              {'user_type': pygame_gui.UI_TEXT_ENTRY_FINISHED,
                                                               'text': self.text,
                                                               'ui_element': self,
                                                               'ui_object_id': self.object_ids[-1]})
                    pygame.event.post(entry_finished_event)
                    processed_event = True
                elif event.key == pygame.K_a and event.mod & pygame.KMOD_CTRL:
                    self.select_range = [0, len(self.text)]
                    self.edit_position = len(self.text)
                    self.cursor_has_moved_recently = True
                    processed_event = True
                elif event.key == pygame.K_x and event.mod & pygame.KMOD_CTRL:
                    if abs(self.select_range[0] - self.select_range[1]) > 0:
                        low_end = min(self.select_range[0], self.select_range[1])
                        high_end = max(self.select_range[0], self.select_range[1])
                        clipboard_copy(self.text[low_end:high_end])
                        self.text = self.text[:low_end] + self.text[high_end:]
                        self.edit_position = low_end
                        self.select_range = [0, 0]
                        self.cursor_has_moved_recently = True
                        processed_event = True
                elif event.key == pygame.K_c and event.mod & pygame.KMOD_CTRL:
                    if abs(self.select_range[0] - self.select_range[1]) > 0:
                        low_end = min(self.select_range[0], self.select_range[1])
                        high_end = max(self.select_range[0], self.select_range[1])
                        clipboard_copy(self.text[low_end:high_end])
                        processed_event = True
                elif event.key == pygame.K_v and event.mod & pygame.KMOD_CTRL:
                    new_text = clipboard_paste()
                    if self.validate_text_string(new_text):
                        if abs(self.select_range[0] - self.select_range[1]) > 0:
                            low_end = min(self.select_range[0], self.select_range[1])
                            high_end = max(self.select_range[0], self.select_range[1])
                            final_text = self.text[:low_end] + new_text + self.text[high_end:]
                            within_length_limit = True
                            if self.length_limit is not None:
                                if len(final_text) > self.length_limit:
                                    within_length_limit = False
                            if within_length_limit:
                                self.text = final_text
                                self.edit_position = low_end + len(new_text)
                                self.select_range = [0, 0]
                                self.cursor_has_moved_recently = True
                        elif len(new_text) > 0:
                            final_text = self.text[:self.edit_position] + new_text + self.text[self.edit_position:]
                            within_length_limit = True
                            if self.length_limit is not None:
                                if len(final_text) > self.length_limit:
                                    within_length_limit = False
                            if within_length_limit:
                                self.text = final_text
                                self.edit_position += len(new_text)
                                self.cursor_has_moved_recently = True
                        processed_event = True
                elif event.key == pygame.K_BACKSPACE:
                    if abs(self.select_range[0] - self.select_range[1]) > 0:
                        low_end = min(self.select_range[0], self.select_range[1])
                        high_end = max(self.select_range[0], self.select_range[1])
                        self.text = self.text[:low_end] + self.text[high_end:]
                        self.edit_position = low_end
                        self.select_range = [0, 0]
                        self.cursor_has_moved_recently = True
                    elif self.edit_position > 0:
                        if self.start_text_offset > 0:
                            self.start_text_offset -= self.font.size(self.text[self.edit_position-1])[0]
                        self.text = self.text[:self.edit_position-1] + self.text[self.edit_position:]
                        self.edit_position -= 1
                        self.cursor_has_moved_recently = True
                    processed_event = True
                elif event.key == pygame.K_DELETE:
                    if abs(self.select_range[0] - self.select_range[1]) > 0:
                        low_end = min(self.select_range[0], self.select_range[1])
                        high_end = max(self.select_range[0], self.select_range[1])
                        self.text = self.text[:low_end] + self.text[high_end:]
                        self.edit_position = low_end
                        self.select_range = [0, 0]
                        self.cursor_has_moved_recently = True
                    elif self.edit_position < len(self.text):
                        self.text = self.text[:self.edit_position] + self.text[self.edit_position+1:]
                        self.edit_position = self.edit_position
                        self.cursor_has_moved_recently = True
                    processed_event = True
                elif event.key == pygame.K_LEFT:
                    if abs(self.select_range[0] - self.select_range[1]) > 0:
                        self.edit_position = min(self.select_range[0], self.select_range[1])
                        self.select_range = [0, 0]
                        self.cursor_has_moved_recently = True
                    elif self.edit_position > 0:
                        self.edit_position -= 1
                        self.cursor_has_moved_recently = True
                    processed_event = True
                elif event.key == pygame.K_RIGHT:
                    if abs(self.select_range[0] - self.select_range[1]) > 0:
                        self.edit_position = max(self.select_range[0], self.select_range[1])
                        self.select_range = [0, 0]
                        self.cursor_has_moved_recently = True
                    elif self.edit_position < len(self.text):
                        self.edit_position += 1
                        self.cursor_has_moved_recently = True
                    processed_event = True
                else:
                    within_length_limit = True
                    if self.length_limit is not None:
                        if len(self.text) >= self.length_limit:
                            within_length_limit = False
                    if within_length_limit:
                        character = event.unicode
                        char_metrics = self.font.metrics(character)
                        if len(char_metrics) > 0 and char_metrics[0] is not None:
                            valid_character = True
                            if self.allowed_characters is not None:
                                if character not in self.allowed_characters:
                                    valid_character = False
                            if self.forbidden_characters is not None:
                                if character in self.forbidden_characters:
                                    valid_character = False
                            if valid_character:
                                if abs(self.select_range[0] - self.select_range[1]) > 0:
                                    low_end = min(self.select_range[0], self.select_range[1])
                                    high_end = max(self.select_range[0], self.select_range[1])
                                    self.text = self.text[:low_end] + character + self.text[high_end:]
                                    self.edit_position = low_end + 1
                                    self.select_range = [0, 0]
                                else:
                                    start_str = self.text[:self.edit_position]
                                    end_str = self.text[self.edit_position:]
                                    self.text = start_str + character + end_str
                                    self.edit_position += 1
                                self.cursor_has_moved_recently = True
                                processed_event = True
        return processed_event

    def find_edit_position_from_pixel_pos(self, pixel_pos: int):
        """
        Locates the correct position to move the edit cursor to, when reacting to a mouse click inside the text entry
        element.

        :param pixel_pos: The x position of our click after being adjusted for text in our box scrolling off-screen.
        """
        start_pos = (self.rect.x + self.border_width + self.shadow_width +
                     self.shape_corner_radius + self.horiz_line_padding)
        acc_pos = start_pos
        index = 0
        for char in self.text:
            x_width = self.font.size(char)[0]

            if acc_pos + (x_width/2) > pixel_pos:
                break
            else:
                index += 1
            acc_pos += x_width
        return index

    def set_allowed_characters(self, allowed_characters: Union[str, List[str]]):
        """
        Sets a whitelist of characters that will be the only ones allowed in our text entry element.
        We can either set the list directly, or request one of the already existing lists by a string
        identifier. The currently supported lists for allowed characters are:

        - 'numbers'

        :param allowed_characters: The characters to allow, either in a list form or one of the supported string ids.
        """
        if type(allowed_characters) is str:
            if allowed_characters == 'numbers':
                self.allowed_characters = UITextEntryLine._number_character_set
            else:
                warnings.warn('Trying to set allowed characters by type string, but no match: '
                              'did you mean to use a list?')

        else:
            self.allowed_characters = allowed_characters.copy()

    def set_forbidden_characters(self, forbidden_characters: Union[str, List[str]]):
        """
        Sets a blacklist of characters that will be banned from our text entry element.
        We can either set the list directly, or request one of the already existing lists by a string
        identifier. The currently supported lists for forbidden characters are:

        - 'numbers'
        - 'forbidden_file_path'

        :param forbidden_characters: The characters to forbid, either in a list form or one of the supported string ids.
        """
        if type(forbidden_characters) is str:
            if forbidden_characters == 'numbers':
                self.forbidden_characters = UITextEntryLine._number_character_set
            elif forbidden_characters == 'forbidden_file_path':
                self.forbidden_characters = UITextEntryLine._forbidden_file_path_characters
            else:
                warnings.warn('Trying to set forbidden characters by type string, but no match: '
                              'did you mean to use a list?')

        else:
            self.forbidden_characters = forbidden_characters.copy()

    def validate_text_string(self, text_to_validate: str) -> bool:
        """
        Checks a string of text to see if any of it's characters don't meet the requirements of the allowed and
        forbidden character sets.

        :param text_to_validate: The text string to check.
        """
        is_valid = True
        if self.forbidden_characters is not None:
            for character in text_to_validate:
                if character in self.forbidden_characters:
                    is_valid = False

        if is_valid and self.allowed_characters is not None:
            for character in text_to_validate:
                if character not in self.allowed_characters:
                    is_valid = False

        return is_valid

    def rebuild_from_changed_theme_data(self):
        """
        Called by the UIManager to check the theming data and rebuild whatever needs rebuilding for this element when
        the theme data has changed.
        """
        has_any_changed = False

        font = self.ui_theme.get_font(self.object_ids, self.element_ids)
        if font != self.font:
            self.font = font
            has_any_changed = True

        shape_type = 'rectangle'
        shape_type_string = self.ui_theme.get_misc_data(self.object_ids, self.element_ids, 'shape')
        if shape_type_string is not None:
            if shape_type_string in ['rectangle', 'rounded_rectangle']:
                shape_type = shape_type_string
        if shape_type != self.shape_type:
            self.shape_type = shape_type
            has_any_changed = True

        corner_radius = 2
        shape_corner_radius_string = self.ui_theme.get_misc_data(self.object_ids,
                                                                 self.element_ids, 'shape_corner_radius')
        if shape_corner_radius_string is not None:
            try:
                corner_radius = int(shape_corner_radius_string)
            except ValueError:
                corner_radius = 2
        if corner_radius != self.shape_corner_radius:
            self.shape_corner_radius = corner_radius
            has_any_changed = True

        border_width = 1
        border_width_string = self.ui_theme.get_misc_data(self.object_ids, self.element_ids, 'border_width')
        if border_width_string is not None:
            try:
                border_width = int(border_width_string)
            except ValueError:
                border_width = 1

        if border_width != self.border_width:
            self.border_width = border_width
            has_any_changed = True

        shadow_width = 2
        shadow_width_string = self.ui_theme.get_misc_data(self.object_ids, self.element_ids, 'shadow_width')
        if shadow_width_string is not None:
            try:
                shadow_width = int(shadow_width_string)
            except ValueError:
                shadow_width = 2
        if shadow_width != self.shadow_width:
            self.shadow_width = shadow_width
            has_any_changed = True

        padding_str = self.ui_theme.get_misc_data(self.object_ids, self.element_ids, 'padding')
        if padding_str is None:
            horiz_line_padding = 2
            vert_line_padding = 2
        else:
            try:
                horiz_line_padding = int(padding_str.split(',')[0])
                vert_line_padding = int(padding_str.split(',')[1])
            except ValueError:
                horiz_line_padding = 2
                vert_line_padding = 2

        if horiz_line_padding != self.horiz_line_padding or vert_line_padding != self.vert_line_padding:
            self.vert_line_padding = vert_line_padding
            self.horiz_line_padding = horiz_line_padding
            has_any_changed = True

        background_colour = self.ui_theme.get_colour_or_gradient(self.object_ids, self.element_ids, 'dark_bg')
        if background_colour != self.background_colour:
            self.background_colour = background_colour
            has_any_changed = True

        border_colour = self.ui_theme.get_colour_or_gradient(self.object_ids, self.element_ids, 'normal_border')
        if border_colour != self.border_colour:
            self.border_colour = border_colour
            has_any_changed = True

        text_colour = self.ui_theme.get_colour_or_gradient(self.object_ids, self.element_ids, 'normal_text')
        if text_colour != self.text_colour:
            self.text_colour = text_colour
            has_any_changed = True

        selected_text_colour = self.ui_theme.get_colour_or_gradient(self.object_ids, self.element_ids, 'selected_text')
        if selected_text_colour != self.selected_text_colour:
            self.selected_text_colour = selected_text_colour
            has_any_changed = True

        selected_bg_colour = self.ui_theme.get_colour_or_gradient(self.object_ids, self.element_ids, 'selected_bg')
        if selected_bg_colour != self.selected_bg_colour:
            self.selected_bg_colour = selected_bg_colour
            has_any_changed = True

        if has_any_changed:
            self.rebuild()



import warnings
import html.parser

import pygame


class CharStyle:
    def __init__(self, char_style=None):
        if char_style is not None:
            self.bold = char_style.bold
            self.italic = char_style.italic
            self.underline = char_style.underline
        else:
            self.bold = False
            self.italic = False
            self.underline = False

    def __eq__(self, other):
        return self.bold == other.bold and self.italic == other.italic and self.underline == other.underline


class TextLineContext:
    """
    A class that covers all the states of the text 'options' so we know what
    to apply when rendering a letter.
    """
    def __init__(self, font_size, font_name, style, color, bg_color, is_link, link_href):
        self.font_size = font_size
        self.font_name = font_name
        self.style = style
        self.color = color
        self.bg_color = bg_color
        self.is_link = is_link
        self.link_href = link_href

    def __eq__(self, other):
        font_size_eq = self.font_size == other.font_size
        font_name_eq = self.font_name == other.font_name
        font_style_eq = self.style == other.style
        font_color_eq = self.color == other.color
        font_bg_color_eq = self.bg_color == other.bg_color
        link_eq = self.is_link == other.is_link
        link_href_eq = self.link_href == other.link_href
        if font_size_eq and font_name_eq and font_style_eq and\
                font_color_eq and font_bg_color_eq and link_eq and link_href_eq:
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


class TextStyleData:
    default_style = {
        'font_name': 'fira_code',
        'font_size': 14
    }

    def __init__(self, theme, element_ids, object_id):
        super().__init__()

        self.ui_theme = theme
        self.len_text = 0
        self.element_stack = []
        self.style_stack = []
        self.current_style = {}
        self.next_style = {}
        self.text_data = ""

        self.indexed_styles = {}

        self.char_style = CharStyle()

        font_info = self.ui_theme.get_font_info(object_id, element_ids)

        self.default_font_name = font_info['name']
        self.default_font_size = int(font_info['size'])
        self.default_style['font_name'] = self.default_font_name
        self.default_style['font_size'] = self.default_font_size

        self.default_font_color = self.ui_theme.get_colour_or_gradient(object_id, element_ids, 'normal_text')
        self.default_bg_color = self.ui_theme.get_colour_or_gradient(object_id, element_ids, 'dark_bg')
        self.font_name = self.default_font_name
        self.font_size = self.default_font_size
        self.font_color = self.default_font_color
        self.bg_color = self.default_bg_color

        self.is_link = False
        self.link_href = ''

        self.push_style('default_style', self.default_style)

    def push_style(self, key, styles):
        old_styles = {}
        for name in styles.keys():
            old_styles[name] = self.current_style.get(name)
        self.style_stack.append((key, old_styles))
        self.current_style.update(styles)
        self.next_style.update(styles)

    def pop_style(self, key):
        # Don't do anything if key is not in stack
        for match, _ in self.style_stack:
            if key == match:
                break
        else:
            return

        # Remove all innermost elements until key is closed.
        while True:
            match, old_styles = self.style_stack.pop()
            self.next_style.update(old_styles)
            self.current_style.update(old_styles)
            if match == key:
                break

    def add_text(self, text):
        self.add_indexed_style(len(self.text_data))
        self.text_data = self.text_data[:self.len_text] + text + self.text_data[self.len_text:]
        self.len_text += len(text)

    def add_indexed_style(self, index):
        if 'bold' in self.current_style.keys():
            self.char_style.bold = self.current_style['bold']
            if self.char_style.bold is None:
                self.char_style.bold = False
        else:
            self.char_style.bold = False
        if 'italic' in self.current_style.keys():
            self.char_style.italic = self.current_style['italic']
            if self.char_style.italic is None:
                self.char_style.italic = False
        else:
            self.char_style.italic = False
        if 'underline' in self.current_style.keys():
            self.char_style.underline = self.current_style['underline']
            if self.char_style.underline is None:
                self.char_style.underline = False
        else:
            self.char_style.underline = False
        if 'font_name' in self.current_style.keys():
            self.font_name = self.current_style['font_name']
            if self.font_name is None:
                self.font_name = self.default_font_name
        else:
            self.font_name = self.default_font_name
        if 'font_size' in self.current_style.keys():
            self.font_size = self.current_style['font_size']
            if self.font_size is None:
                self.font_size = self.default_font_size
        else:
            self.font_size = self.default_font_size
        if 'font_color' in self.current_style.keys():
            self.font_color = self.current_style['font_color']
            if self.font_color is None:
                self.font_color = self.default_font_color
        else:
            self.font_color = self.default_font_color
        if 'bg_color' in self.current_style.keys():
            self.bg_color = self.current_style['bg_color']
            if self.bg_color is None:
                self.bg_color = self.default_bg_color
        else:
            self.bg_color = self.default_bg_color

        if 'link' in self.current_style.keys():
            self.is_link = self.current_style['link']
            if self.is_link is None:
                self.is_link = False
        else:
            self.is_link = False

        if 'link_href' in self.current_style.keys():
            self.link_href = self.current_style['link_href']
            if self.link_href is None:
                self.link_href = ''
        else:
            self.link_href = ''

        self.indexed_styles[index] = TextLineContext(self.font_size,
                                                     self.font_name,
                                                     CharStyle(self.char_style),
                                                     self.font_color,
                                                     self.bg_color,
                                                     self.is_link,
                                                     self.link_href)


class TextHTMLParser(TextStyleData, html.parser.HTMLParser):
    font_sizes = {
        1: 8,
        1.5: 9,
        2: 10,
        2.5: 11,
        3: 12,
        3.5: 13,
        4: 14,
        4.5: 16,
        5: 18,
        5.5: 20,
        6: 24,
        6.5: 32,
        7: 48
    }

    def __init__(self, ui_theme, element_ids, object_id):
        super().__init__(ui_theme, element_ids, object_id)
        self.ui_theme = ui_theme
        self.element_ids = element_ids
        self.object_id = object_id

    def handle_starttag(self, tag, attrs):
        element = tag.lower()

        self.element_stack.append(element)

        attributes = {}
        for key, value in attrs:
            attributes[key.lower()] = value

        style = {}
        if element in ('b', 'strong'):
            style['bold'] = True
        elif element == 'a':
            style['link'] = True
            if 'href' in attributes:
                style["link_href"] = attributes['href']
        elif element in ('i', 'em', 'var'):
            style['italic'] = True
        elif element == 'u':
            style['underline'] = True
        elif element == 'font':
            if 'face' in attributes:
                font_name = attributes['face']  # .split(',')
                style["font_name"] = font_name
            if 'size' in attributes:
                font_size = self.font_sizes[float(attributes['size'])]
                style["font_size"] = font_size
            if 'color' in attributes:
                if attributes['color'][0] == '#':
                    style["font_color"] = pygame.color.Color(attributes['color'])
                else:
                    style["font_color"] = self.ui_theme.get_colour_or_gradient(self.object_id,
                                                                               self.element_ids, attributes['color'])

        elif element == 'body':
            if 'bgcolor' in attributes:
                if attributes['bgcolor'][0] == '#':
                    style["bg_color"] = pygame.color.Color(attributes['bgcolor'])
                else:
                    style["bg_color"] = self.ui_theme.get_colour_or_gradient(self.object_id,
                                                                             self.element_ids, attributes['bg_color'])
        elif element == 'br':
            self.add_text('\n')  # u'\u2028'
        else:
            warning_text = 'Unsupported HTML Tag <' + element + '>. Check documentation' \
                                                                ' for full range of supported tags.'
            warnings.warn(warning_text, UserWarning)

        self.push_style(element, style)

    def handle_endtag(self, tag):
        element = tag.lower()

        if element not in self.element_stack:
            return

        self.pop_style(element)

        while self.element_stack.pop() != element:
            pass

    def handle_data(self, data):
        self.add_text(data)

    def error(self, message):
        pass



import pygame
from typing import List, Union, Tuple

from pygame_gui.core import ui_container
from pygame_gui import ui_manager


class UIElement(pygame.sprite.Sprite):
    """
    A base class for UI elements. You shouldn't create UI Element objects, instead all UI Element classes should
    derive from this class.

    Inherits from pygame.sprite.Sprite.

    :param relative_rect: A rectangle shape of the UI element, the position is relative to the element's container.
    :param manager: The UIManager that manages this UIElement.
    :param container: A container that this element is contained in.
    :param starting_height: Used to record how many layers above it's container this element should be. Normally 1.
    :param layer_thickness: Used to record how 'thick' this element is in layers. Normally 1.
    :param element_ids: A list of ids that describe the 'hierarchy' of UIElements that this UIElement is part of.
    :param object_ids: A list of custom defined IDs that describe the 'hierarchy' that this UIElement is part of.
    """
    def __init__(self, relative_rect: pygame.Rect, manager: 'ui_manager.UIManager',
                 container: Union['ui_container.UIContainer', None],
                 starting_height: int, layer_thickness: int,
                 object_ids: Union[List[Union[str, None]], None] = None,
                 element_ids: Union[List[str], None] = None):

        self._layer = 0
        self.ui_manager = manager
        super().__init__(self.ui_manager.get_sprite_group())
        self.relative_rect = relative_rect
        self.ui_group = self.ui_manager.get_sprite_group()
        self.ui_theme = self.ui_manager.get_theme()
        self.object_ids = object_ids
        self.element_ids = element_ids

        self.layer_thickness = layer_thickness
        self.starting_height = starting_height
        self.top_layer = self._layer + self.layer_thickness

        if container is None:
            root_window = self.ui_manager.get_window_stack().get_root_window()
            if root_window is not None:
                container = root_window.get_container()
            else:
                container = self

        self.ui_container = container
        if self.ui_container is not None and self.ui_container is not self:
            self.ui_container.add_element(self)

        if self.ui_container is self:
            self.rect = pygame.Rect((relative_rect.x,
                                    relative_rect.y),
                                    relative_rect.size)
        else:
            self.rect = pygame.Rect((self.ui_container.rect.x + relative_rect.x,
                                     self.ui_container.rect.y + relative_rect.y),
                                    relative_rect.size)

        self.image = None

        self.is_enabled = True
        self.hovered = False
        self.hover_time = 0.0

    @staticmethod
    def create_valid_ids(parent_element, object_id, element_id):
        """
        Creates valid id lists for an element. It will assert if users supply object IDs that won't work such as those
        containing full stops. These ID lists are used by the theming system to identify what theming parameters to
        apply to which element.

        :param parent_element: Element that this element 'belongs to' in theming. Elements inherit colours from parents.
        :param object_id: An optional ID to help distinguish this element from other elements of the same class.
        :param element_id: A string ID representing this element's class.
        :return:
        """
        if object_id is not None and ('.' in object_id or ' ' in object_id):
            raise ValueError('Object ID cannot contain fullstops or spaces: ' + str(object_id))

        if parent_element is not None:
            new_element_ids = parent_element.element_ids.copy()
            new_element_ids.append(element_id)

            new_object_ids = parent_element.object_ids.copy()
            new_object_ids.append(object_id)
        else:
            new_element_ids = [element_id]
            new_object_ids = [object_id]

        return new_element_ids, new_object_ids

    def update_containing_rect_position(self):
        """
        Updates the position of this element based on the position of it's container. Usually called when the container
        has moved.
        """
        self.rect = pygame.Rect((self.ui_container.rect.x + self.relative_rect.x,
                                 self.ui_container.rect.y + self.relative_rect.y),
                                self.relative_rect.size)

    def change_layer(self, new_layer: int):
        """
        Changes the layer this element is on.

        :param new_layer: The layer to change this element to.
        """
        self.ui_group.change_layer(self, new_layer)
        self._layer = new_layer
        self.top_layer = new_layer + self.layer_thickness

    def kill(self):
        """
        Overriding regular sprite kill() method to remove the element from it's container.
        """
        self.ui_container.remove_element(self)
        super().kill()

    def check_hover(self, time_delta: float, hovered_higher_element: bool) -> bool:
        """
        A method that helps us to determine which, if any, UI Element is currently being hovered by the mouse.

        :param time_delta: A float, the time in seconds between the last call to this function and now (roughly).
        :param hovered_higher_element: A boolean, representing whether we have already hovered a 'higher' element.
        :return bool: A boolean that is true if we have hovered a UI element, either just now or before this method.
        """
        if self.alive() and self.can_hover():
            mouse_x, mouse_y = pygame.mouse.get_pos()
            mouse_pos = pygame.math.Vector2(mouse_x, mouse_y)

            if self.is_enabled and self.hover_point(mouse_x, mouse_y) and not hovered_higher_element:
                if not self.hovered:
                    self.hovered = True
                    self.on_hovered()

                hovered_higher_element = True
                self.while_hovering(time_delta, mouse_pos)

            else:
                if self.hovered:
                    self.hovered = False
                    self.on_unhovered()
        elif self.hovered:
            self.hovered = False
        return hovered_higher_element

    def set_relative_position(self, position: Union[pygame.math.Vector2, Tuple[int, int], Tuple[float, float]]):
        """
        Method to directly set the relative rect position of a button.

        :param position: The new position to set.
        """
        self.relative_rect.x = int(position[0])
        self.relative_rect.y = int(position[1])
        self.rect.x = self.relative_rect.x + self.ui_container.rect.x
        self.rect.y = self.relative_rect.y + self.ui_container.rect.y

    def set_position(self, position: Union[pygame.math.Vector2, Tuple[int, int], Tuple[float, float]]):
        """
        Method to directly set the absolute rect position of a button.

        :param position: The new position to set.
        """
        self.rect.x = int(position[0])
        self.rect.y = int(position[1])
        self.relative_rect.x = self.rect.x - self.ui_container.rect.x
        self.relative_rect.y = self.rect.y - self.ui_container.rect.y

    def on_hovered(self):
        """
        A stub to override. Called when this UI element first enters the 'hovered' state.
        """
        pass

    def while_hovering(self, time_delta: float, mouse_pos: pygame.math.Vector2):
        """
        A stub method to override. Called when this UI element is currently hovered.

        :param time_delta: A float, the time in seconds between the last call to this function and now (roughly).
        :param mouse_pos: The current position of the mouse as 2D Vector.
        """
        pass

    def hover_point(self, x: float, y: float) -> bool:
        """
        Test if a given point counts as 'hovering' this UI element. Normally that is a straightforward matter of
        seeing if a point is inside the rectangle. Occasionally it will also check if we are in a wider zone around
        a UI element once it is already active, this makes it easier to move scroll bars and the like.

        :param x: The x (horizontal) position of the point.
        :param y: The y (vertical) position of the point.
        :return bool: Returns True if we are hovering this element.
        """
        return self.rect.collidepoint(x, y)

    def on_unhovered(self):
        """
        A stub to override. Called when this UI element leaves the 'hovered' state.
        """
        pass

    def can_hover(self) -> bool:
        """
        A stub method to override. Called to test if this method can be hovered.
        """
        return True

    def process_event(self, event: pygame.event.Event) -> bool:
        """
        A stub to override. Gives UI Elements access to pygame events.

        :param event: The event to process.
        :return bool: Should return True if this element makes use fo this event.
        """
        if self is not None:
            return False

    def select(self):
        """
        A stub to override. Called when we select focus this UI element.
        """
        pass

    def unselect(self):
        """
        A stub to override. Called when we stop select focusing this UI element.
        """
        pass

    def rebuild_from_changed_theme_data(self):
        pass



"""
Some  constants used across the GUI
TODO: Using strings for now for compatibility but we could deprecate these to be integers later.
"""

# UI Event types
UI_BUTTON_PRESSED = 'ui_button_pressed'
UI_TEXT_BOX_LINK_CLICKED = 'ui_text_box_link_clicked'
UI_TEXT_ENTRY_FINISHED = 'ui_text_entry_finished'
UI_DROP_DOWN_MENU_CHANGED = 'ui_drop_down_menu_changed'

# Text effects
TEXT_EFFECT_TYPING_APPEAR = 'typing_appear'
TEXT_EFFECT_FADE_IN = 'fade_in'
TEXT_EFFECT_FADE_OUT = 'fade_out'



import pygame
from typing import Tuple, Union
import warnings


class ShadowGenerator:
    """
    A class to generate surfaces that work as a 'shadow' for rectangular UI elements. Base shadow surface are generated
    with an algorithm, then when one is requested at a specific size the closest pre-generated shadow surface is
    picked and then scaled to the exact size requested.

    By default it creates a four base shadows in a small range of sizes. If you find the shadow appearance
    unsatisfactory then it is possible to create more closer to the size of the elements you are having trouble with.
    """

    def __init__(self):
        self.created_rectangle_shadows = {}
        self.created_ellipse_shadows = {}
        self.preloaded_shadow_corners = {}

    def shadow_edges_and_corners_to_preload(self, list_of_corners, list_of_edges):
        # corners = {corner_radius: 10, shadow_width: 10}
        pass

    def create_shadow_corners(self, shadow_width_param, corner_radius_param, aa_amount=4):

        if shadow_width_param <= 0:
            shadow_width_param = 1
            warnings.warn("Tried to make shadow with width <= 0")
        final_corner_surface = pygame.Surface((corner_radius_param * aa_amount,
                                               corner_radius_param * aa_amount),
                                              flags=pygame.SRCALPHA, depth=32)
        final_corner_surface.fill(pygame.Color('#00000000'))

        final_edge_surface = pygame.Surface((shadow_width_param * aa_amount,
                                             shadow_width_param * aa_amount),
                                            flags=pygame.SRCALPHA, depth=32)
        final_edge_surface.fill(pygame.Color('#00000000'))

        alpha_increment = 20.0 / (shadow_width_param ** 1.5)
        shadow_alpha = alpha_increment

        corner_shadow_width = corner_radius_param * aa_amount
        corner_shadow_height = corner_radius_param * aa_amount
        corner_radius = corner_radius_param * aa_amount
        corner_center = (corner_radius, corner_radius)

        edge_shadow_width = shadow_width_param * aa_amount
        edge_shadow_height = shadow_width_param * aa_amount
        edge_shadow_fade_height = edge_shadow_width

        for i in range(0, shadow_width_param):
            if corner_shadow_width > 0 and corner_shadow_height > 0 and corner_radius > 0:
                # Edge
                edge_shadow_rect = pygame.Rect(0, 0, edge_shadow_width, edge_shadow_height)
                edge_shadow_surface = pygame.Surface(edge_shadow_rect.size, flags=pygame.SRCALPHA, depth=32)
                edge_shadow_surface.fill(pygame.Color('#00000000'))

                edge_shadow_surface.fill(pygame.Color(0, 0, 0, int(shadow_alpha)),
                                         pygame.Rect(0, edge_shadow_height - edge_shadow_fade_height,
                                                     edge_shadow_width,
                                                     edge_shadow_fade_height))

                final_edge_surface.blit(edge_shadow_surface, (0, 0), special_flags=pygame.BLEND_RGBA_ADD)

                # corner
                corner_shadow_rect = pygame.Rect(0, 0, corner_shadow_width, corner_shadow_height)
                corner_shadow_surface = pygame.Surface(corner_shadow_rect.size, flags=pygame.SRCALPHA, depth=32)
                corner_shadow_surface.fill(pygame.Color('#00000000'))

                pygame.draw.circle(corner_shadow_surface, pygame.Color(0, 0, 0, int(shadow_alpha)),
                                   corner_center,
                                   corner_radius)

                final_corner_surface.blit(corner_shadow_surface, (0, 0), special_flags=pygame.BLEND_RGBA_ADD)

                # increments/decrements
                shadow_alpha += alpha_increment
                corner_radius -= aa_amount
                edge_shadow_fade_height -= aa_amount

        top_edge = pygame.transform.smoothscale(final_edge_surface, (shadow_width_param, shadow_width_param))
        bottom_edge = pygame.transform.flip(top_edge, False, True)
        left_edge = pygame.transform.rotate(top_edge, 90)
        right_edge = pygame.transform.flip(left_edge, True, False)

        tl_corner = pygame.transform.smoothscale(final_corner_surface, (corner_radius_param, corner_radius_param))
        tr_corner = pygame.transform.flip(tl_corner, True, False)
        bl_corner = pygame.transform.flip(tl_corner, False, True)
        br_corner = pygame.transform.flip(tl_corner, True, True)

        corners_and_edges = {"top": top_edge,
                             "bottom": bottom_edge,
                             "left": left_edge,
                             "right": right_edge,
                             "top_left": tl_corner,
                             "top_right": tr_corner,
                             "bottom_left": bl_corner,
                             "bottom_right": br_corner}
        self.preloaded_shadow_corners[str(shadow_width_param) + 'x' + str(corner_radius_param)] = corners_and_edges

        return corners_and_edges

    def create_new_rectangle_shadow_new(self, width: int, height: int,
                                        shadow_width_param: int, corner_radius_param: int, ):

        if width >= corner_radius_param and height >= corner_radius_param:
            final_surface = pygame.Surface((width, height), flags=pygame.SRCALPHA, depth=32)

            corner_index_id = str(shadow_width_param) + 'x' + str(corner_radius_param)
            if corner_index_id in self.preloaded_shadow_corners:
                edges_and_corners = self.preloaded_shadow_corners[corner_index_id]
            else:
                edges_and_corners = self.create_shadow_corners(shadow_width_param, corner_radius_param)

            final_surface.blit(edges_and_corners["top_left"], (0, 0))
            final_surface.blit(edges_and_corners["top_right"], (width - corner_radius_param, 0))

            final_surface.blit(edges_and_corners["bottom_left"],
                               (0, height - corner_radius_param))
            final_surface.blit(edges_and_corners["bottom_right"],
                               (width - corner_radius_param, height - corner_radius_param))

            if width - (2 * corner_radius_param) > 0:
                top_edge = pygame.transform.scale(edges_and_corners["top"],
                                                  (width - (2 * corner_radius_param), shadow_width_param))
                bottom_edge = pygame.transform.scale(edges_and_corners["bottom"],
                                                     (width - (2 * corner_radius_param), shadow_width_param))
                final_surface.blit(top_edge, (corner_radius_param, 0))
                final_surface.blit(bottom_edge, (corner_radius_param, height - shadow_width_param))

            if height - (2 * corner_radius_param) > 0:
                left_edge = pygame.transform.scale(edges_and_corners["left"],
                                                   (shadow_width_param, height - (2 * corner_radius_param)))
                right_edge = pygame.transform.scale(edges_and_corners["right"],
                                                    (shadow_width_param, height - (2 * corner_radius_param)))

                final_surface.blit(left_edge, (0, corner_radius_param))
                final_surface.blit(right_edge, (width - shadow_width_param, corner_radius_param))

            self.created_rectangle_shadows[str(width) + 'x' + str(height) +
                                           'x' + str(corner_radius_param) + 'x' + str(shadow_width_param)] = final_surface
            return final_surface
        else:
            return None

    def create_new_rectangle_shadow_old(self, width: int, height: int,
                                        shadow_width_param: int, corner_radius_param: int, aa_amount: int = 4):
        """
        Creates a rectangular shadow surface at the specified size and stores it for later use.

        :param width: The width of the base shadow to create.
        :param height: The height of the base shadow to create.
        :param shadow_width_param: The width of the shadowed edge.
        :param aa_amount: The amount of anti-aliasing to use, defaults to 4.
        :param corner_radius_param: The radius of the rectangular shadow's corners.
        """
        print("creating new shadow, width:" + str(width) +
              ", height:" + str(height) + ", shadow_width: " + str(shadow_width_param))

        final_shadow_surface = pygame.Surface((width * aa_amount, height * aa_amount), flags=pygame.SRCALPHA, depth=32)
        final_shadow_surface.fill(pygame.Color('#00000000'))

        alpha_increment = 20.0 / (shadow_width_param ** 1.5)
        shadow_alpha = alpha_increment
        shadow_width = width * aa_amount
        shadow_height = height * aa_amount
        corner_radius = corner_radius_param * aa_amount
        for i in range(0, shadow_width_param):
            if shadow_width > 0 and shadow_height > 0:
                shadow_rect = pygame.Rect(i * aa_amount, i * aa_amount, shadow_width, shadow_height)
                shadow_surface = pygame.Surface(shadow_rect.size, flags=pygame.SRCALPHA, depth=32)
                shadow_surface.fill(pygame.Color('#00000000'))

                pygame.draw.circle(shadow_surface, pygame.Color(0, 0, 0, int(shadow_alpha)),
                                   (corner_radius,
                                    corner_radius), corner_radius)
                shadow_surface.fill(pygame.Color('#00000000'),
                                    pygame.Rect(corner_radius, 0,
                                                corner_radius,
                                                corner_radius * 2))
                shadow_surface.fill(pygame.Color('#00000000'),
                                    pygame.Rect(0, corner_radius,
                                                corner_radius * 2,
                                                corner_radius))

                x_flip = pygame.transform.flip(shadow_surface, True, False)
                shadow_surface.blit(x_flip, (0, 0))
                y_flip = pygame.transform.flip(shadow_surface, False, True)
                shadow_surface.blit(y_flip, (0, 0))

                shadow_surface.fill(pygame.Color(0, 0, 0, int(shadow_alpha)),
                                    pygame.Rect((corner_radius, 0),
                                                (shadow_rect.width - (2 * corner_radius),
                                                 shadow_rect.height)))
                shadow_surface.fill(pygame.Color(0, 0, 0, int(shadow_alpha)),
                                    pygame.Rect((0, corner_radius),
                                                (shadow_rect.width,
                                                 shadow_rect.height - (2 * corner_radius))))

                final_shadow_surface.blit(shadow_surface, shadow_rect, special_flags=pygame.BLEND_RGBA_ADD)

                shadow_width -= (2 * aa_amount)
                shadow_height -= (2 * aa_amount)
                corner_radius -= aa_amount
                corner_radius = max(corner_radius, 2)
                shadow_alpha += alpha_increment

        final_surface = pygame.transform.smoothscale(final_shadow_surface, (width, height))
        self.created_rectangle_shadows[str(width) + 'x' + str(height) +
                                       'x' + str(corner_radius_param) + 'x' + str(shadow_width_param)] = final_surface
        return final_surface

    def create_new_ellipse_shadow(self, width: int, height: int, shadow_width_param: int, aa_amount: int = 4):
        """
        Creates a ellipse shaped shadow surface at the specified size and stores it for later use.

        :param width: The width of the shadow to create.
        :param height: The height of the shadow to create.
        :param shadow_width_param: The width of the shadowed edge.
        :param aa_amount: The amount of anti-aliasing to use, defaults to 4.
        """
        shadow_surface = pygame.Surface((width * aa_amount, height * aa_amount), flags=pygame.SRCALPHA, depth=32)
        shadow_surface.fill(pygame.Color('#00000000'))

        alpha_increment = max(1, int(20 / shadow_width_param))
        shadow_alpha = alpha_increment
        shadow_width = width * aa_amount
        shadow_height = height * aa_amount
        for i in range(0, shadow_width_param):
            if shadow_width > 0 and shadow_height > 0:
                shadow_rect = pygame.Rect(i * aa_amount, i * aa_amount, shadow_width, shadow_height)
                pygame.draw.ellipse(shadow_surface, pygame.Color(0, 0, 0, shadow_alpha), shadow_rect)
                shadow_width -= (2 * aa_amount)
                shadow_height -= (2 * aa_amount)
                shadow_alpha += alpha_increment

        final_surface = pygame.transform.smoothscale(shadow_surface, (width, height))
        self.created_ellipse_shadows[str(width) + 'x' + str(height) + 'x' + str(shadow_width_param)] = final_surface
        return final_surface

    def find_closest_shadow_scale_to_size(self, size: Tuple[int, int],
                                          shadow_width: int = 2,
                                          shape: str = "rectangle",
                                          corner_radius: int = 2) -> Union[pygame.Surface, None]:
        """
        This function searches through our dictionary of created shadows, grabs the closest one to the size
        we request and then scales that shadow to the exact size we need.



        :param size: The size of the element we are finding a shadow for.
        :param shadow_width: The width of the shadow to find.
        :param shape: The shape of the shadow to find.
        :param corner_radius: The radius of the corners if this is a rectangular shadow.
        :return: The shadow surface we asked for scaled to the size we requested, or None if no shadows exist.
        """
        lowest_diff = 1000000000000
        closest_key = None
        if shape == 'rectangle':
            for key in self.created_rectangle_shadows.keys():
                dimension_strings = key.split('x')
                width = int(dimension_strings[0])
                height = int(dimension_strings[1])
                corner_rad = int(dimension_strings[2])
                shadow_size = int(dimension_strings[3])

                key_aspect_ratio = float(width / height)
                request_aspect_ratio = float(size[0] / size[1])
                ratio_diff = abs(key_aspect_ratio - request_aspect_ratio)
                width_diff = abs(width - size[0])
                height_diff = abs(height - size[1])
                corner_diff = abs(corner_rad - corner_radius)
                shadow_diff = abs(shadow_size - shadow_width)
                total_diff = width_diff + height_diff + shadow_diff + corner_diff + ratio_diff
                if corner_diff < 2 and shadow_diff < 2 and ratio_diff < 0.75 and total_diff < lowest_diff:
                    lowest_diff = total_diff
                    closest_key = key

            if closest_key is not None:
                return pygame.transform.smoothscale(self.created_rectangle_shadows[closest_key], size)
            else:
                return self.create_new_rectangle_shadow_new(size[0], size[1], shadow_width, corner_radius)

        if shape == 'ellipse':
            for key in self.created_ellipse_shadows.keys():
                dimension_strings = key.split('x')
                width = int(dimension_strings[0])
                height = int(dimension_strings[1])
                shadow_size = int(dimension_strings[2])

                width_diff = abs(width - size[0])
                height_diff = abs(height - size[1])
                ratio_diff = max(width_diff, height_diff)
                shadow_diff = abs(shadow_size - shadow_width) * 50
                total_diff = width_diff + height_diff + shadow_diff
                if total_diff < lowest_diff:
                    lowest_diff = total_diff
                    if ratio_diff < 10:
                        closest_key = key

            if closest_key is not None:
                return pygame.transform.smoothscale(self.created_ellipse_shadows[closest_key], size)
            else:
                return self.create_new_ellipse_shadow(size[0], size[1], shadow_width)

        return None



from pygame_gui.ui_manager import UIManager
from pygame_gui import core
from pygame_gui import elements
from pygame_gui import windows
from pygame_gui.elements import text
from pygame_gui._constants import *



import os
import io
import base64
import pygame
import warnings


# Only import the 'stringified' data if we can't find the actual default font file
# This is need for a working PyInstaller build
ROOT_PATH = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
FONT_PATH = os.path.normpath(os.path.join(ROOT_PATH, 'data/FiraCode-Regular.ttf'))
if not os.path.exists(FONT_PATH):
    from pygame_gui.core._string_data import FiraCode_Regular, FiraCode_Bold
    from pygame_gui.core._string_data import FiraMono_BoldItalic, FiraMono_RegularItalic


class UIFontDictionary:
    """
    The font dictionary is used to store all the fonts that have been loaded into the UI system.
    """
    _html_font_sizes = {
        1: 8,
        1.5: 9,
        2: 10,
        2.5: 11,
        3: 12,
        3.5: 13,
        4: 14,
        4.5: 16,
        5: 18,
        5.5: 20,
        6: 24,
        6.5: 32,
        7: 48
    }

    _html_font_sizes_reverse_lookup = {
        8: 1,
        9: 1.5,
        10: 2,
        11: 2.5,
        12: 3,
        13: 3.5,
        14: 4,
        16: 4.5,
        18: 5,
        20: 5.5,
        24: 6,
        32: 6.5,
        48: 7
    }

    def __init__(self):
        self.default_font_size = 14

        # Only use  the 'stringified' data if we can't find the actual default font file
        # This is need for a working PyInstaller build
        module_root_path = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
        default_font_file_path = os.path.normpath(os.path.join(module_root_path, 'data/FiraCode-Regular.ttf'))
        if os.path.exists(default_font_file_path):
            self.loaded_fonts = {'fira_code_regular_14': pygame.font.Font(default_font_file_path,
                                                                          self.default_font_size)}

            self.known_font_paths = {'fira_code': [os.path.normpath(os.path.join(module_root_path,
                                                                                 'data/FiraCode-Regular.ttf')),
                                                   os.path.normpath(os.path.join(module_root_path,
                                                                                 'data/FiraCode-Bold.ttf')),
                                                   os.path.normpath(os.path.join(module_root_path,
                                                                                 'data/FiraMono-RegularItalic.ttf')),
                                                   os.path.normpath(os.path.join(module_root_path,
                                                                                 'data/FiraMono-BoldItalic.ttf'))]}
        else:
            fira_code_regular_file_object = io.BytesIO(base64.standard_b64decode(FiraCode_Regular))
            self.loaded_fonts = {'fira_code_regular_14': pygame.font.Font(fira_code_regular_file_object,
                                                                          self.default_font_size)}

            self.known_font_paths = {'fira_code': [FiraCode_Regular,
                                                   FiraCode_Bold,
                                                   FiraMono_RegularItalic,
                                                   FiraMono_BoldItalic]}

        self.used_font_ids = ['fira_code_regular_14']

    def find_font(self, font_size: int, font_name: str,
                  bold: bool = False, italic: bool = False) -> pygame.font.Font:
        """
        Find a loaded font from the font dictionary. Will load a font if it does not already exist and we have paths
        to the needed files, however it will issue a warning after doing so because dynamic file loading is normally a
        bad idea as you will get frame rate hitches while the running program waits for the font to load.

        Instead it's best to preload all your needed files at another time in your program when you have more control
        over the user experience.

        :param font_size: The size of the font to find.
        :param font_name: The name of the font to find.
        :param bold: Whether the font is bold or not.
        :param italic: Whether the font is italic or not.
        :return pygame.font.Font: Returns either the font we asked for, or the default font.
        """
        font_id = self.create_font_id(font_size, font_name, bold, italic)

        if font_id not in self.used_font_ids:
            self.used_font_ids.append(font_id)  # record font usage for optimisation purposes

        if font_id in self.loaded_fonts:  # font already loaded
            return self.loaded_fonts[font_id]
        elif font_name in self.known_font_paths:  # we know paths to this font, just haven't loaded current size/style

            style_string = "regular"
            if bold and italic:
                style_string = "bold_italic"
            elif bold:
                style_string = "bold"
            elif italic:
                style_string = "italic"

            warning_string = ('Finding font with id: ' + font_id + " that is not already loaded.\n"
                                                                   "Preload this font with {'name': "
                                                                   "'" + font_name + "',"
                                                                   " 'point_size': " + str(font_size) + ","
                                                                   " 'style': '" + style_string + "'}")
            warnings.warn(warning_string, UserWarning)
            self.preload_font(font_size, font_name, bold, italic)
            return self.loaded_fonts[font_id]
        else:
            return self.loaded_fonts["default_font_regular_12"]

    @staticmethod
    def create_font_id(font_size: int, font_name: str, bold: bool, italic: bool) -> str:
        """
        Create an id for a particularly styled and sized font from those characteristics.

        :param font_size: The size of the font.
        :param font_name: The name of the font.
        :param bold: Whether the font is bold styled or not.
        :param italic: Whether the font is italic styled or not.
        :return str: The finished font id.
        """
        if bold and italic:
            font_style_string = "bold_italic"
        elif bold and not italic:
            font_style_string = "bold"
        elif not bold and italic:
            font_style_string = "italic"
        else:
            font_style_string = "regular"
        font_id = font_name + "_" + font_style_string + "_" + str(font_size)
        return font_id

    def preload_font(self, font_size: int, font_name: str, bold: bool = False, italic: bool = False):
        """
        Lets us load a font at a particular size and style before we use it. While you can get away with relying on
        dynamic font loading during development, it is better to eventually pre-load all your font data at a
        controlled time, which is where this method comes in.

        :param font_size: The size of the font to load.
        :param font_name: The name of the font to load.
        :param bold: Whether the font is bold styled or not.
        :param italic: Whether the font is italic styled or not.
        """
        font_id = self.create_font_id(font_size, font_name, bold, italic)
        if font_id in self.loaded_fonts:  # font already loaded
            warnings.warn('Trying to pre-load font id: ' + font_id + ' that is already loaded', UserWarning)
        elif font_name in self.known_font_paths:  # we know paths to this font, just haven't loaded current size/style
            if bold and italic:
                try:
                    if type(self.known_font_paths[font_name][3]) == bytes:
                        file_loc = io.BytesIO(base64.standard_b64decode(self.known_font_paths[font_name][3]))
                    else:
                        file_loc = self.known_font_paths[font_name][3]
                    new_font = pygame.font.Font(file_loc, font_size)
                    new_font.set_bold(True)
                    new_font.set_italic(True)
                    self.loaded_fonts[font_id] = new_font
                except FileNotFoundError:
                    warnings.warn("Failed to load font at path: " + self.known_font_paths[font_name][3])

            elif bold and not italic:
                try:
                    if type(self.known_font_paths[font_name][1]) == bytes:
                        file_loc = io.BytesIO(base64.standard_b64decode(self.known_font_paths[font_name][1]))
                    else:
                        file_loc = self.known_font_paths[font_name][1]
                    new_font = pygame.font.Font(file_loc, font_size)
                    new_font.set_bold(True)
                    self.loaded_fonts[font_id] = new_font
                except FileNotFoundError:
                    warnings.warn("Failed to load font at path: " + self.known_font_paths[font_name][1])
            elif not bold and italic:
                try:
                    if type(self.known_font_paths[font_name][2]) == bytes:
                        file_loc = io.BytesIO(base64.standard_b64decode(self.known_font_paths[font_name][2]))
                    else:
                        file_loc = self.known_font_paths[font_name][2]
                    new_font = pygame.font.Font(file_loc, font_size)
                    new_font.set_italic(True)
                    self.loaded_fonts[font_id] = new_font
                except FileNotFoundError:
                    warnings.warn("Failed to load font at path: " + self.known_font_paths[font_name][2])
            else:
                try:
                    if type(self.known_font_paths[font_name][0]) == bytes:
                        file_loc = io.BytesIO(base64.standard_b64decode(self.known_font_paths[font_name][0]))
                    else:
                        file_loc = self.known_font_paths[font_name][0]
                    new_font = pygame.font.Font(file_loc, font_size)
                    self.loaded_fonts[font_id] = new_font
                except FileNotFoundError:
                    warnings.warn("Failed to load font at path: " + self.known_font_paths[font_name][0])
        else:
            raise UserWarning('Trying to pre-load font id:' + font_id + 'with no paths set')

    def add_font_path(self, font_name: str, font_path: str, bold_path: str = None,
                      italic_path: str = None, bold_italic_path: str = None):
        """
        Adds paths to different font files for a font name.

        :param font_name: The name to assign to these font files.
        :param font_path: The path to the font's file with no particular style.
        :param bold_path: The path to the font's file with a bold style.
        :param italic_path: The path to the font's file with an italic style.
        :param bold_italic_path: The path to the font's file with a bold and an italic style.
        """
        if font_name not in self.known_font_paths:
            if bold_path is None:
                bold_path = font_path
            if italic_path is None:
                italic_path = font_path
            if bold_italic_path is None:
                bold_italic_path = font_path
            self.known_font_paths[font_name] = [os.path.normpath(font_path),
                                                os.path.normpath(bold_path),
                                                os.path.normpath(italic_path),
                                                os.path.normpath(bold_italic_path)]

    def print_unused_loaded_fonts(self):
        """
        Can be called to check if the UI is loading any fonts that we haven't used by the point this function is
        called. If a font is truly unused then we can remove it from our loading and potentially speed up the overall
        loading of the program.

        This is not a foolproof check because this function could easily be called before we have explored all the code
        paths in a project that may use fonts.
        """
        unused_font_ids = []
        for key in self.loaded_fonts.keys():
            if key not in self.used_font_ids:
                unused_font_ids.append(key)

        if len(unused_font_ids) > 0:
            print('Unused font ids:')
            for font_id in unused_font_ids:
                html_size = UIFontDictionary._html_font_sizes_reverse_lookup[int(font_id.split('_')[-1])]
                print(font_id + '(HTML size: ' + str(html_size) + ')')

    def convert_html_size_to_point_size(self, html_size: float) -> int:
        """
        Takes in a HTML style font size and converts it into a point font size.

        :param html_size:
        :return int: A 'point' font size we can use with pygame.font
        """
        if html_size in UIFontDictionary._html_font_sizes:
            return UIFontDictionary._html_font_sizes[html_size]
        else:
            return self.default_font_size



from pygame_gui.core.ui_appearance_theme import UIAppearanceTheme
from pygame_gui.core.ui_container import UIContainer
from pygame_gui.core.ui_element import UIElement
from pygame_gui.core.ui_font_dictionary import UIFontDictionary
from pygame_gui.core.ui_shadow import ShadowGenerator
from pygame_gui.core.ui_window import UIWindow
from pygame_gui.core.ui_window_stack import UIWindowStack



import pygame
from typing import Union

from pygame_gui import ui_manager
from pygame_gui.core import ui_container
from pygame_gui.core.ui_element import UIElement


class UIImage(UIElement):
    """
    Displays a pygame surface as a UI element, intended for an image but it can serve other purposes.

    :param relative_rect: The rectangle that contains and positions the image relative to it's container.
    :param image_surface:
    :param manager: The UIManager that manages this element.
    :param container: The container that this element is within. If set to None will be the root window's container.
    :param parent_element: The element this element 'belongs to' in the theming hierarchy.
    :param object_id: A custom defined ID for fine tuning of theming.
    """
    def __init__(self, relative_rect: pygame.Rect, image_surface: pygame.Surface,
                 manager: ui_manager.UIManager,
                 container: ui_container.UIContainer = None,
                 parent_element: UIElement = None,
                 object_id: Union[str, None] = None):

        new_element_ids, new_object_ids = self.create_valid_ids(parent_element=parent_element,
                                                                object_id=object_id,
                                                                element_id='image')

        super().__init__(relative_rect, manager, container,
                         starting_height=1,
                         layer_thickness=1,
                         object_ids=new_object_ids,
                         element_ids=new_element_ids)
        self.image = image_surface



from typing import Union

from pygame_gui.core import ui_window


class UIWindowStack:
    """
    A class for managing a stack of GUI windows so that only one is 'in front' at a time and the rest are sorted based
    on the last time they were interacted with/created.
    """
    def __init__(self, window_resolution):
        self.window_resolution = window_resolution
        self.stack = []

    def clear(self):
        """
        Empties the whole stack removing and killing all windows.
        """
        while len(self.stack) != 0:
            self.stack.pop().kill()
        self.stack.clear()

    def add_new_window(self, window: ui_window.UIWindow):
        """
        Adds a window to the top of the stack.

        :param window: The window to add.
        """
        if len(self.stack) > 0:
            new_layer = self.stack[-1].get_top_layer() + 1
        else:
            new_layer = 0
        window.change_window_layer(new_layer)
        self.stack.append(window)

    def remove_window(self, window_to_remove: ui_window.UIWindow):
        """
        Removes a window from the stack and resorts the remaining windows to adjust for it's absence.

        :param window_to_remove: the window to remove.
        """
        if window_to_remove in self.stack:
            popped_windows_to_add_back = []
            window = self.stack.pop()
            while window != window_to_remove:
                popped_windows_to_add_back.append(window)
                window = self.stack.pop()

            popped_windows_to_add_back.reverse()
            for old_window in popped_windows_to_add_back:
                self.add_new_window(old_window)

    def move_window_to_front(self, window_to_front: ui_window.UIWindow):
        """
        Moves the passed in window to the top of the window stack and resorts the other windows to deal with the
        change.

        :param window_to_front: the window to move to the front.
        """
        if window_to_front in self.stack:
            popped_windows_to_add_back = []
            window = self.stack.pop()
            while window != window_to_front:
                popped_windows_to_add_back.append(window)
                window = self.stack.pop()

            popped_windows_to_add_back.reverse()
            for old_window in popped_windows_to_add_back:
                self.add_new_window(old_window)

            self.add_new_window(window_to_front)

    def get_root_window(self) -> Union[ui_window.UIWindow, None]:
        """
        Gets the 'root' window, which should always be the first one in the stack and should represent an imaginary
        window the size of the whole pygame application's display surface.

        :return Union[ui_window.UIWindow, None]:  The 'root' window
        """
        if len(self.stack) > 0:
            return self.stack[0]
        else:
            return None

    def is_window_at_top(self, window: ui_window.UIWindow) -> bool:
        """
        Checks if a window is at the top of the window stack or not.

        :param window: The window to check.
        :return bool: returns True if this window is at the top of the stack.
        """
        if window is self.stack[-1]:
            return True
        else:
            return False



import pygame
from typing import Tuple, List, Dict, Union

from pygame_gui.core.ui_appearance_theme import UIAppearanceTheme
from pygame_gui.core.ui_window_stack import UIWindowStack
from pygame_gui.core.ui_window import UIWindow
from pygame_gui.core.ui_element import UIElement


class UIManager:
    """
    The UI Manager class helps keep track of all the moving parts in the pygame_gui system.

    Before doing anything else with pygame_gui create a UIManager and remember to update it every frame.
    """

    def __init__(self, window_resolution: Tuple[int, int], theme_path: str = None, enable_live_theme_updates=True):
        self.window_resolution = window_resolution
        self.ui_theme = UIAppearanceTheme()
        if theme_path is not None:
            self.ui_theme.load_theme(theme_path)
        self.ui_group = pygame.sprite.LayeredUpdates()

        self.select_focused_element = None
        self.last_focused_vertical_scrollbar = None

        self.ui_window_stack = UIWindowStack(self.window_resolution)
        UIWindow(pygame.Rect((0, 0), self.window_resolution), self, ['root_window'])

        self.live_theme_updates = enable_live_theme_updates
        self.theme_update_acc = 0.0
        self.theme_update_check_interval = 1.0

    def get_theme(self) -> UIAppearanceTheme:
        """
        Gets the theme so the data in it can be accessed.
        :return: UIAppearanceTheme - the theme data used by this UIManager
        """
        return self.ui_theme

    def get_sprite_group(self) -> pygame.sprite.LayeredUpdates:
        """
        Gets the sprite group used by the entire UI to keep it in the correct order for drawing and processing input.
        :return : pygame.sprite.LayeredUpdates -
        """
        return self.ui_group

    def get_window_stack(self) -> UIWindowStack:
        """
        The UIWindowStack organises any windows in the UI Manager so that they are correctly sorted and move windows
        we interact with to the top of the stack.
        :return: UIWindowStack
        """
        return self.ui_window_stack

    def get_shadow(self, size: Tuple[int, int], shadow_width: int = 2,
                   shape: str = 'rectangle', corner_radius: int = 2) -> pygame.Surface:
        """
        Returns a 'shadow' surface scaled to the requested size.


        :param size: The size of the object we are shadowing + it's shadow.
        :param shadow_width: The width of the shadowed edge.
        :param shape: The shape of the requested shadow
        :param corner_radius: The radius of the shadow corners if this is a rectangular shadow.
        :return: A shadow.
        """
        return self.ui_theme.shadow_generator.find_closest_shadow_scale_to_size(size, shadow_width,
                                                                                shape, corner_radius)

    def set_window_resolution(self, window_resolution):
        self.window_resolution = window_resolution

    def clear_and_reset(self):
        """
        Clear all existing windows, including the root window, which should get rid of all created UI elements.
        We then recreate the UIWindowStack and the root window.
        """
        self.ui_window_stack.clear()
        self.ui_window_stack = UIWindowStack(self.window_resolution)
        UIWindow(pygame.Rect((0, 0), self.window_resolution), self, ['root_window'])

    def process_events(self, event: pygame.event.Event):
        """
        This is the top level method through which all input to UI elements is processed and reacted to.

        One of the key things it controls is the currently 'focused' or 'selected' element of which there
        can be only one at a time.

        TODO: Pass 'clicked inside' checks down to elements so we can properly test if clicks are inside shapes or not.

        :param event:  pygame.event.Event - the event to process.
        """
        event_handled = False
        window_sorting_event_handled = False
        sorted_layers = sorted(self.ui_group.layers(), reverse=True)
        for layer in sorted_layers:
            sprites_in_layer = self.ui_group.get_sprites_from_layer(layer)
            # a little bit of code to pop windows we start messing with to the front
            if not window_sorting_event_handled:
                windows_in_layer = [window for window in sprites_in_layer if (
                        'window' in window.element_ids[-1]) and (self.ui_window_stack.get_root_window() is not window)]
                for window in windows_in_layer:
                    if not window_sorting_event_handled:
                        window_sorting_event_handled = window.check_clicked_inside(event)
            if not event_handled:
                for ui_element in sprites_in_layer:
                    if not event_handled:
                        if event.type == pygame.MOUSEBUTTONDOWN:
                            if event.button == 1:
                                mouse_x, mouse_y = event.pos
                                if ui_element.rect.collidepoint(mouse_x, mouse_y):
                                    if ui_element is not self.select_focused_element:
                                        self.unselect_focus_element()
                                        self.select_focus_element(ui_element)

                        event_handled = ui_element.process_event(event)

    def update(self, time_delta: float):
        """
        From here all our UI elements are updated and which element is currently 'hovered' is checked; which means
        the mouse pointer is overlapping them. This is managed centrally so we aren't ever overlapping two elements at
        once.

        It also updates the shape cache to continue storing already created elements shapes in the long term cache, in
        case we need them later.

        Finally, if live theme updates are enabled, it checks to see if the theme file has been modified and triggers
        all the UI elements to rebuild if it has.

        :param time_delta: The time passed since the last call to update, in seconds.
        """

        if self.live_theme_updates:
            self.theme_update_acc += time_delta
            if self.theme_update_acc > self.theme_update_check_interval:
                self.theme_update_acc = 0.0
                if self.ui_theme.check_need_to_reload():
                    for sprite in self.ui_group.sprites():
                        sprite.rebuild_from_changed_theme_data()

        self.ui_theme.update_shape_cache()

        hover_handled = False
        sorted_layers = sorted(self.ui_group.layers(), reverse=True)
        for layer in sorted_layers:
            for ui_element in self.ui_group.get_sprites_from_layer(layer):
                element_handled_hover = ui_element.check_hover(time_delta, hover_handled)
                if element_handled_hover:
                    hover_handled = True

        self.ui_group.update(time_delta)

    def draw_ui(self, window_surface: pygame.Surface):
        """
        Draws all the UI elements on the screen. Generally you want this to be after the rest of your game sprites
        have been drawn.

        If you want to do something particularly unusual with drawing you may have to write your own UI manager.

        :param window_surface: The screen or window surface on which we are going to draw all of our UI Elements.

        """
        self.ui_group.draw(window_surface)

    def add_font_paths(self, font_name: str, regular_path: str, bold_path: str = None,
                       italic_path: str = None, bold_italic_path: str = None):
        """
        Add file paths for custom fonts you want to use in the UI. For each font name you add you can specify
        font files for different styles. Fonts with designed styles tend to render a lot better than fonts that are
        forced to make use of pygame's bold and italic styling effects, so if you plan to use bold and italic text
        at small sizes - find fonts with these styles available as separate files.

        The font name you specify here can be used to choose the font in the blocks of HTML-subset formatted text,
        available in some of the UI elements like the UITextBox.

        It is recommended that you also pre-load any fonts you use at an appropriate moment in your project
        rather than letting the library dynamically load them when they are required. That is because dynamic loading
        of large font files can cause UI elements with a lot of font usage to appear rather slowly as they are waiting
        for the fonts they need to load.

        :param font_name: The name of the font that will be used to reference it elsewhere in the GUI.
        :param regular_path: The path of the font file for this font with no styles applied.
        :param bold_path: The path of the font file for this font with just bold style applied.
        :param italic_path: The path of the font file for this font with just italic style applied.
        :param bold_italic_path: The path of the font file for this font with bold & italic style applied.

        """
        self.get_theme().get_font_dictionary().add_font_path(font_name,
                                                             regular_path,
                                                             bold_path,
                                                             italic_path,
                                                             bold_italic_path)

    def preload_fonts(self, font_list: List[Dict[str, Union[str, int, float]]]):
        """
        It's a good idea to pre-load the exact fonts your program uses during the loading phase of your program.
        By default the pygame_gui library will still work, but will spit out reminder warnings when you haven't
        done this. Loading fonts on the fly will slow down the apparent responsiveness when creating UI elements
        that use a lot of different fonts.

        To pre-load custom fonts, or to use custom fonts at all (i.e. ones that aren't the default 'fira_code' font)
        you must first add the paths to the files for those fonts, then load the specific fonts with a list of font
        descriptions in a dictionary form like so:

        code:`{'name': 'fira_code', 'point_size': 12, 'style': 'bold_italic'}`

        You can specify size either in pygame.Font point sizes with 'point_size', or in HTML style sizes with
        'html_size'. Style options are:

        - 'regular'
        - 'italic'
        - 'bold'
        - 'bold_italic'

        The name parameter here must match the one you used when you added the file paths.

        :param font_list: A list of font descriptions in dictionary format as described above.

        """
        for font in font_list:
            name = 'fira_code'
            bold = False
            italic = False
            size = 14
            if 'name' in font:
                name = font['name']
            if 'style' in font:
                if 'bold' in font['style']:
                    bold = True
                if 'italic' in font['style']:
                    italic = True
            if 'html_size' in font:
                size = self.ui_theme.get_font_dictionary().convert_html_size_to_point_size(font['html_size'])
            elif 'point_size' in font:
                size = font['point_size']

            self.ui_theme.get_font_dictionary().preload_font(size, name, bold, italic)

    def print_unused_fonts(self):
        """
        Helps you identify which pre-loaded fonts you are actually still using in your project after you've
        fiddled around with the text a lot by printing out a list of fonts that have not been used yet at the
        time this function is called.

        Of course if you don't run the code path in which a particular font is used before calling this function
        then it won't be of much use, so take it's results under advisement rather than as gospel.

        """
        self.ui_theme.get_font_dictionary().print_unused_loaded_fonts()

    def unselect_focus_element(self):
        """
        Unselect and clear the currently focused element.
        """
        if self.select_focused_element is not None:
            self.select_focused_element.unselect()
            self.select_focused_element = None

    def select_focus_element(self, ui_element: UIElement):
        """
        Set an element as the focused element.

        If the element is a scroll bar we also keep track of that.

        :param ui_element: The element to focus on.
        """
        if self.select_focused_element is None:
            self.select_focused_element = ui_element
            self.select_focused_element.select()

            if 'vertical_scroll_bar' in ui_element.element_ids:
                self.last_focused_vertical_scrollbar = ui_element

    def clear_last_focused_from_vert_scrollbar(self, vert_scrollbar):
        """
        Clears the last scrollbar that we used.

        :param vert_scrollbar: A scrollbar UIElement.
        """
        if vert_scrollbar is self.last_focused_vertical_scrollbar:
            self.last_focused_vertical_scrollbar = None

    def get_last_focused_vert_scrollbar(self):
        """
        Gets the last scrollbar that we used.

        :return: A UIElement.
        """
        return self.last_focused_vertical_scrollbar



from pygame_gui.elements.text.html_parser import TextHTMLParser, CharStyle, TextLineContext
from pygame_gui.elements.text.text_effects import TextBoxEffect, TypingAppearEffect, FadeOutEffect, FadeInEffect



import pygame
import json
import os
import io
import base64
import warnings
from contextlib import contextmanager

from typing import Union

from pygame_gui.core.ui_font_dictionary import UIFontDictionary
from pygame_gui.core.ui_shadow import ShadowGenerator

# Only import the 'stringified' data if we can't find the actual default theme file
# This is need for working PyInstaller build
ROOT_PATH = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
THEME_PATH = os.path.normpath(os.path.join(ROOT_PATH, 'data/default_theme.json'))
if not os.path.exists(THEME_PATH):
    from pygame_gui.core._string_data import default_theme

try:
    from os import PathLike  # for Python 3.6
except ImportError:
    PathLike = None


class UIAppearanceTheme:
    """
    The Appearance Theme class handles all the data that styles and generally dictates the appearance of UI elements
    across the whole UI.

    The styling is split into four general areas:

    - colours - spelled in the British English fashion with a 'u'.
    - font - specifying a font to use for a UIElement where that is a relevant consideration.
    - images - describing any images to be used in a UIElement.
    - misc - covering all other types of data and stored as strings.

    To change the theming for the UI you normally specify a theme file when creating the UIManager. For more
    information on theme files see the specific documentation elsewhere.
    """
    def __init__(self):

        # the base colours are the default colours all UI elements use if they
        # don't have a more specific colour defined for their element
        self.base_colours = {'normal_bg': pygame.Color('#25292e'),
                             'hovered_bg': pygame.Color('#35393e'),
                             'disabled_bg': pygame.Color('#25292e'),
                             'selected_bg': pygame.Color('#193754'),
                             'active_bg': pygame.Color('#193754'),
                             'dark_bg': pygame.Color('#15191e'),
                             'normal_text': pygame.Color('#c5cbd8'),
                             'hovered_text': pygame.Color('#FFFFFF'),
                             'selected_text': pygame.Color('#FFFFFF'),
                             'active_text': pygame.Color('#FFFFFF'),
                             'disabled_text': pygame.Color('#6d736f'),
                             'normal_border': pygame.Color('#DDDDDD'),
                             'hovered_border': pygame.Color('#EDEDED'),
                             'disabled_border': pygame.Color('#909090'),
                             'selected_border': pygame.Color('#294764'),
                             'active_border': pygame.Color('#294764'),
                             'link_text': pygame.Color('#c5cbFF'),
                             'link_hover': pygame.Color('#a5abDF'),
                             'link_selected': pygame.Color('#DFabDF'),
                             'text_shadow': pygame.Color('#777777'),
                             'filled_bar': pygame.Color("#f4251b"),
                             'unfilled_bar': pygame.Color("#CCCCCC")}

        # colours for specific elements stored by element id then colour id
        self.ui_element_colours = {}

        # font dictionary that stores loaded fonts
        self.font_dictionary = UIFontDictionary()

        # shadow generator
        self.shadow_generator = ShadowGenerator()

        # shape cache
        self.shape_cache = ShapeCache()

        # the font to use if no other font is specified
        # these hardcoded paths should be OK for PyInstaller right now because they will never actually used while
        # fira_code is the default pre-loaded font. May need to re-visit this later.
        module_root_path = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
        self.base_font_info = {'name': 'fira_code',
                               'size': 14,
                               'bold': False,
                               'italic': False,
                               'regular_path': os.path.normpath(os.path.join(module_root_path,
                                                                             'data/FiraCode-Regular.ttf')),
                               'bold_path': os.path.normpath(os.path.join(module_root_path,
                                                                          'data/FiraCode-Bold.ttf')),
                               'italic_path': os.path.normpath(os.path.join(module_root_path,
                                                                            'data/FiraMono-RegularItalic.ttf')),
                               'bold_italic_path': os.path.normpath(os.path.join(module_root_path,
                                                                                 'data/FiraMono-BoldItalic.ttf'))}

        # fonts for specific elements stored by element id
        self.ui_element_fonts_info = {}
        self.ui_element_fonts = {}

        # stores any images specified in themes that need loading at the appropriate time
        self.ui_element_image_paths = {}
        self.ui_element_image_surfaces = {}
        self.loaded_image_files = {}  # just a dictionary of all images paths & image files loaded by the UI

        # stores everything that doesn't have a specific place elsewhere and doesn't need any time-consuming loading
        # all will be stored as strings and will have to do any further processing in their specific elements.
        # misc data that doesn't have a value defined in a theme will return None so elements should be prepared
        # to handle that with a default behaviour
        self.ui_element_misc_data = {}

        self._theme_file_last_modified = None
        self._theme_file_path = None

        # Only load  the 'stringified' data if we can't find the actual default theme file
        # This is need for PyInstaller build
        default_theme_file_path = os.path.normpath(os.path.join(module_root_path, 'data','default_theme.json'))
        if os.path.exists(default_theme_file_path):
            self.load_theme(default_theme_file_path)
        else:
            default_theme_file = io.BytesIO(base64.standard_b64decode(default_theme))
            self.load_theme(default_theme_file)

    def get_font_dictionary(self):
        """
        Lets us grab the Font dictionary, which is created by the theme object, so we can access it directly.

        :return UIFontDictionary:
        """
        return self.font_dictionary

    def check_need_to_reload(self):
        """
        Check if we need to reload our theme file because it's been modified. If so, trigger a reload and return True
        so that the UIManager can trigger elements to rebuild from the theme data.

        :return bool: True if we need to reload elements because the theme data has changed.
        """
        if self._theme_file_path is not None:
            try:
                stamp = os.stat(self._theme_file_path).st_mtime
            except FileNotFoundError:
                return False

            if stamp != self._theme_file_last_modified:
                self._theme_file_last_modified = stamp
                self.reload_theming()
                return True
            else:
                return False

    def update_shape_cache(self):
        self.shape_cache.update()

    def reload_theming(self):
        """
        We need to load our theme file see if anything expensive has changed, if so trigger it to reload/rebuild.

        """
        self.load_theme(self._theme_file_path)

    def load_fonts(self):
        """
        Loads all fonts specified in our loaded theme.

        """
        self.font_dictionary.add_font_path(self.base_font_info['name'],
                                           self.base_font_info['regular_path'],
                                           self.base_font_info['bold_path'],
                                           self.base_font_info['italic_path'],
                                           self.base_font_info['bold_italic_path'])

        font_id = self.font_dictionary.create_font_id(self.base_font_info['size'],
                                                      self.base_font_info['name'],
                                                      self.base_font_info['bold'],
                                                      self.base_font_info['italic'])

        if font_id not in self.font_dictionary.loaded_fonts:
            self.font_dictionary.preload_font(self.base_font_info['size'],
                                              self.base_font_info['name'],
                                              self.base_font_info['bold'],
                                              self.base_font_info['italic'])

        for element_key in self.ui_element_fonts_info.keys():
            font_info = self.ui_element_fonts_info[element_key]

            bold_path = None
            italic_path = None
            bold_italic_path = None
            if 'regular_path' in font_info:
                regular_path = font_info['regular_path']

                if 'bold_path' in font_info:
                    bold_path = font_info['bold_path']
                if 'italic_path' in font_info:
                    italic_path = font_info['italic_path']
                if 'bold_italic_path' in font_info:
                    bold_italic_path = font_info['bold_italic_path']

                self.font_dictionary.add_font_path(font_info['name'],
                                                   regular_path,
                                                   bold_path,
                                                   italic_path,
                                                   bold_italic_path)

            font_id = self.font_dictionary.create_font_id(font_info['size'],
                                                          font_info['name'],
                                                          font_info['bold'],
                                                          font_info['italic'])

            if font_id not in self.font_dictionary.loaded_fonts:
                self.font_dictionary.preload_font(font_info['size'],
                                                  font_info['name'],
                                                  font_info['bold'],
                                                  font_info['italic'])

            self.ui_element_fonts[element_key] = self.font_dictionary.find_font(font_info['size'],
                                                                                font_info['name'],
                                                                                font_info['bold'],
                                                                                font_info['italic'])

    def load_images(self):
        """
        Loads all images in our loaded theme.

        """
        for element_key in self.ui_element_image_paths.keys():
            image_paths_dict = self.ui_element_image_paths[element_key]
            if element_key not in self.ui_element_image_surfaces:
                self.ui_element_image_surfaces[element_key] = {}
            for path_key in image_paths_dict:
                if image_paths_dict[path_key]['changed']:
                    path = image_paths_dict[path_key]['path']
                    image = None
                    if path in self.loaded_image_files:
                        image = self.loaded_image_files[path]
                    else:
                        try:
                            image = pygame.image.load(os.path.normpath(path)).convert_alpha()
                            self.loaded_image_files[path] = image
                        except pygame.error:
                            warnings.warn('Unable to load image at path: ' + str(os.path.abspath(path)))

                    if image is not None:
                        if 'sub_surface_rect' in image_paths_dict[path_key]:
                            surface = image.subsurface(image_paths_dict[path_key]['sub_surface_rect'])
                        else:
                            surface = image
                        self.ui_element_image_surfaces[element_key][path_key] = surface

    def preload_shadow_edges(self):
        """
        Looks through the theming data for any shadow edge combos we haven't loaded yet and tries to pre-load them.
        This helps stop the UI from having to create the complicated parts of the shadows dynamically which can be
        noticeably slow (e.g. waiting a second for a window to appear).

        For this to work correctly the theme file shouldn't contain any 'invalid' data that is later clamped by the UI,
        Plus, it is helpful if any rounded rectangles that set a corner radius also set a shadow width at the same time.
        """
        for misc_id in self.ui_element_misc_data:

            shape = 'rectangle'
            shadow_width = 2
            shape_corner_radius = 2

            if 'shape' in self.ui_element_misc_data[misc_id]:
                shape = self.ui_element_misc_data[misc_id]['shape']

            if 'shadow_width' in self.ui_element_misc_data[misc_id]:
                try:
                    shadow_width = int(self.ui_element_misc_data[misc_id]['shadow_width'])
                except ValueError:
                    shadow_width = 2
                    warnings.warn(
                        "Invalid value: " + self.ui_element_misc_data[misc_id]['shadow_width'] +
                        " for shadow_width")

            if 'shape_corner_radius' in self.ui_element_misc_data[misc_id]:
                try:
                    shape_corner_radius = int(self.ui_element_misc_data[misc_id]['shape_corner_radius'])
                except ValueError:
                    shape_corner_radius = 2
                    warnings.warn(
                        "Invalid value: " + self.ui_element_misc_data[misc_id]['shape_corner_radius'] +
                        " for shape_corner_radius")

            # we can preload shadow edges if we are dealing with a rectangular shadow
            if (shape == 'rounded_rectangle' or shape == 'rectangle') and shadow_width > 0:
                if ('shadow_width' in self.ui_element_misc_data[misc_id] and
                        'shape_corner_radius' in self.ui_element_misc_data[misc_id]):
                    shadow_id = str(shadow_width) + 'x' + str(shape_corner_radius)
                    if shadow_id not in self.shadow_generator.preloaded_shadow_corners:
                        self.shadow_generator.create_shadow_corners(shadow_width, shape_corner_radius)
                elif ('shadow_width' in self.ui_element_misc_data[misc_id] and
                      'shape_corner_radius' not in self.ui_element_misc_data[misc_id]):
                    # have a shadow width but no idea on the corners, try most common -
                    shadow_id_1 = str(shadow_width) + 'x' + str(2)
                    if shadow_id_1 not in self.shadow_generator.preloaded_shadow_corners:
                        self.shadow_generator.create_shadow_corners(shadow_width, 2)
                    shadow_id_2 = str(shadow_width) + 'x' + str(shadow_width)
                    if shadow_id_2 not in self.shadow_generator.preloaded_shadow_corners:
                        self.shadow_generator.create_shadow_corners(shadow_width, shadow_width)
                elif ('shape_corner_radius' in self.ui_element_misc_data[misc_id] and
                      'shadow_width' not in self.ui_element_misc_data[misc_id]):
                    # have a corner radius but no idea on the shadow width, try most common -
                    shadow_id_1 = str(1) + 'x' + str(shape_corner_radius)
                    if shadow_id_1 not in self.shadow_generator.preloaded_shadow_corners:
                        self.shadow_generator.create_shadow_corners(1, shape_corner_radius)
                    shadow_id_2 = str(2) + 'x' + str(shape_corner_radius)
                    if shadow_id_2 not in self.shadow_generator.preloaded_shadow_corners:
                        self.shadow_generator.create_shadow_corners(2, shape_corner_radius)
                    shadow_id_3 = str(shape_corner_radius) + 'x' + str(shape_corner_radius)
                    if shadow_id_3 not in self.shadow_generator.preloaded_shadow_corners:
                        self.shadow_generator.create_shadow_corners(shape_corner_radius, shape_corner_radius)

    def get_next_id_node(self, current_node, element_ids, object_ids, index, tree_size, combined_ids):
        if index < tree_size:
            if object_ids is not None:
                if index < len(object_ids):
                    object_id = object_ids[index]
                    if object_id is not None:
                        next_node = {'id': object_id, 'parent': current_node}
                        self.get_next_id_node(next_node, element_ids, object_ids, index + 1, tree_size, combined_ids)
            element_id = element_ids[index]
            next_node_2 = {'id': element_id, 'parent': current_node}
            self.get_next_id_node(next_node_2, element_ids, object_ids, index + 1, tree_size, combined_ids)
        else:
            # unwind
            gathered_ids = []
            unwind_node = current_node
            while unwind_node is not None:
                gathered_ids.append(unwind_node['id'])
                unwind_node = unwind_node['parent']
            gathered_ids.reverse()
            combined_id = gathered_ids[0]
            for index in range(1, len(gathered_ids)):
                combined_id += '.'
                combined_id += gathered_ids[index]
            combined_ids.append(combined_id)

    def build_all_combined_ids(self, element_ids, object_ids):
        """
        Construct a combined element id from the elements ids.

        :param element_ids: All the ids of elements this element is contained within.
        :param object_ids: All the ids of objects this element is contained within.
        :return: The combined id string in the database
        """
        combined_ids = []
        if object_ids is not None and element_ids is not None:
            if len(object_ids) != len(element_ids):
                raise ValueError("Object ID hierarchy is not equal in length to Element ID hierarchy"
                                 "Element IDs: " + str(element_ids) + "\n"
                                 "Object IDs: " + str(object_ids) + "\n")
            if len(element_ids) != 0:
                self.get_next_id_node(None, element_ids, object_ids, 0, len(element_ids), combined_ids)

        return combined_ids

    def get_image(self, object_ids, element_ids, image_id):
        """
        Will return None if no image is specified. There are UI elements that have an optional image display.

        :param image_id: The id identifying the particular image spot in the UI we are looking for an image to add to.
        :param object_ids: A list of custom IDs representing an element's location in a hierarchy.
        :param element_ids: A list of element IDs representing an element's location in a hierarchy.
        :return None or pygame.Surface:
        """

        combined_element_ids = self.build_all_combined_ids(element_ids, object_ids)

        # then check for an element specific data
        for combined_element_id in combined_element_ids:
            if combined_element_id in self.ui_element_image_surfaces:
                if image_id in self.ui_element_image_surfaces[combined_element_id]:
                    return self.ui_element_image_surfaces[combined_element_id][image_id]

        return None

    def get_font_info(self, object_ids, element_ids):
        """
        Uses some data about a UIElement to get font data as dictionary

        :param object_ids: A list of custom IDs representing an element's location in a hierarchy.
        :param element_ids: A list of element IDs representing an element's location in a hierarchy.
        :return dictionary: Data about the font requested
        """
        font_info = self.base_font_info

        # Check for a unique theming for this specific object
        combined_element_ids = self.build_all_combined_ids(element_ids, object_ids)

        for combined_element_id in combined_element_ids:
            if combined_element_id in self.ui_element_fonts_info:
                return self.ui_element_fonts_info[combined_element_id]

        return font_info

    def get_font(self, object_ids, element_ids):
        """
        Uses some data about a UIElement to get a font object.

        :param object_ids: A list of custom IDs representing an element's location in a hierarchy.
        :param element_ids: A list of element IDs representing an element's location in a hierarchy.
        :return pygame.font.Font: A pygame font object.
        """
        # set the default font as the final fall back
        font = self.font_dictionary.find_font(self.base_font_info['size'],
                                              self.base_font_info['name'],
                                              self.base_font_info['bold'],
                                              self.base_font_info['italic'])

        # Check for a unique theming for this specific object
        combined_element_ids = self.build_all_combined_ids(element_ids, object_ids)

        for combined_element_id in combined_element_ids:
            if combined_element_id in self.ui_element_fonts:
                return self.ui_element_fonts[combined_element_id]

        return font

    def get_misc_data(self, object_ids, element_ids, misc_data_id):
        """
        Uses data about a UI element and a specific ID to try and find a piece of miscellaneous theming data.

        :param object_ids: A list of custom IDs representing an element's location in a hierarchy.
        :param element_ids: A list of element IDs representing an element's location in a hierarchy.
        :param misc_data_id: The id for the specific piece of miscellaneous data we are looking for.
        :return None or str: Returns a string if we find the data, otherwise returns None.
        """
        combined_element_ids = self.build_all_combined_ids(element_ids, object_ids)

        # then check for an element specific data
        for combined_element_id in combined_element_ids:
            if combined_element_id in self.ui_element_misc_data:
                if misc_data_id in self.ui_element_misc_data[combined_element_id]:
                    return self.ui_element_misc_data[combined_element_id][misc_data_id]

        return None

    def get_colour(self, object_ids, element_ids, colour_id):
        """
        Uses data about a UI element and a specific ID to find a colour from our theme.

        :param object_ids: A list of custom IDs representing an element's location in a hierarchy.
        :param element_ids: A list of element IDs representing an element's location in a hierarchy.
        :param colour_id: The id for the specific colour we are looking for.
        :return pygame.Color: A pygame colour.
        """
        colour_or_gradient = self.get_colour_or_gradient(object_ids, element_ids, colour_id)
        if type(colour_or_gradient) == ColourGradient:
            gradient = colour_or_gradient
            colour = gradient.colour_1
        elif type(colour_or_gradient) == pygame.Color:
            colour = colour_or_gradient
        else:
            colour = pygame.Color('#000000')
        return colour

    def get_colour_or_gradient(self, object_ids, element_ids, colour_id):
        """
        Uses data about a UI element and a specific ID to find a colour, or a gradient, from our theme.
        Use this function if the UIElement can handle either type.

        :param object_ids: A list of custom IDs representing an element's location in a hierarchy.
        :param element_ids: A list of element IDs representing an element's location in a hierarchy.
        :param colour_id: The id for the specific colour we are looking for.
        :return pygame.Color or ColourGradient: A colour or a gradient object.
        """
        # first check for a unique theming for this specific object
        combined_element_ids = self.build_all_combined_ids(element_ids, object_ids)

        for combined_element_id in combined_element_ids:
            if combined_element_id in self.ui_element_colours:
                if colour_id in self.ui_element_colours[combined_element_id]:
                    return self.ui_element_colours[combined_element_id][colour_id]

        # if we don't have a specific colour for our individual element, try to inherit colours from higher
        # in the hierarchy
        if object_ids is not None:
            for object_id in object_ids:
                if object_id is not None:
                    if object_id in self.ui_element_colours:
                        if colour_id in self.ui_element_colours[object_id]:
                            return self.ui_element_colours[object_id][colour_id]

        if element_ids is not None:
            for element_id in element_ids:
                if element_id in self.ui_element_colours:
                    if colour_id in self.ui_element_colours[element_id]:
                        return self.ui_element_colours[element_id][colour_id]

        # then fall back on default colour with same id
        if colour_id in self.base_colours:
            return self.base_colours[colour_id]

        # if all else fails find a colour with the most similar id words
        colour_parts = colour_id.split('_')
        best_fit_key_count = 0
        best_fit_colour = self.base_colours['normal_bg']
        for key in self.base_colours.keys():
            key_words = key.split('_')
            count = sum(el in colour_parts for el in key_words)
            if count > best_fit_key_count:
                best_fit_key_count = count
                best_fit_colour = self.base_colours[key]
        return best_fit_colour

    @staticmethod
    @contextmanager
    def opened_w_error(filename, mode="r"):
        if type(filename) != io.BytesIO:
            try:
                f = open(filename, mode)
            except IOError as err:
                yield None, err
            else:
                try:
                    yield f, None
                finally:
                    f.close()
        else:
            f = filename
            try:
                yield f, None
            finally:
                f.close()

    def load_theme(self, file_path: Union[str, PathLike, io.BytesIO]):
        """
        Loads a theme file, and currently, all associated data like fonts and images required by the theme.

        :param file_path: The path to the theme we want to load.

        """
        if file_path is None:
            raise ValueError('Theme path cannot be None')

        if type(file_path) != io.BytesIO:
            self._theme_file_path = file_path
            self._theme_file_last_modified = os.stat(self._theme_file_path).st_mtime
            used_file_path = os.path.abspath(file_path)
        else:
            used_file_path = file_path

        with self.opened_w_error(used_file_path, 'r') as (theme_file, error):
            if error:
                warnings.warn("Failed to open theme file at path:" + str(file_path))
                load_success = False
            else:
                try:
                    load_success = True
                    theme_dict = json.load(theme_file)
                except json.decoder.JSONDecodeError:
                    warnings.warn("Failed to load current theme file, check syntax", UserWarning)
                    load_success = False

                if load_success:
                    for element_name in theme_dict.keys():
                        if element_name == 'defaults':
                            self.load_colour_defaults_from_theme(element_name, theme_dict)
                        else:
                            for data_type in theme_dict[element_name]:
                                if data_type == 'font':
                                    self.load_element_font_data_from_theme(data_type, element_name, theme_dict)

                                if data_type == 'colours':
                                    self.load_element_colour_data_from_theme(data_type, element_name, theme_dict)

                                elif data_type == 'images':
                                    self.load_element_image_data_from_theme(data_type, element_name, theme_dict)

                                elif data_type == 'misc':
                                    self.load_element_misc_data_from_theme(data_type, element_name, theme_dict)

        # TODO: these should be triggered at an appropriate time in our project when lots of files are being loaded
        if load_success:
            self.load_fonts()  # save to trigger load with the same data as it won't do anything
            self.load_images()
            self.preload_shadow_edges()

    def load_element_misc_data_from_theme(self, data_type, element_name, theme_dict):
        if element_name not in self.ui_element_misc_data:
            self.ui_element_misc_data[element_name] = {}
        misc_dict = theme_dict[element_name][data_type]
        for misc_data_key in misc_dict:
            self.ui_element_misc_data[element_name][misc_data_key] = str(misc_dict[misc_data_key])

    def load_element_image_data_from_theme(self, data_type, element_name, theme_dict):
        if element_name not in self.ui_element_image_paths:
            self.ui_element_image_paths[element_name] = {}
        images_dict = theme_dict[element_name][data_type]
        for image_key in images_dict:
            if image_key not in self.ui_element_image_paths[element_name]:
                self.ui_element_image_paths[element_name][image_key] = {}
                self.ui_element_image_paths[element_name][image_key]['changed'] = True
            else:
                self.ui_element_image_paths[element_name][image_key]['changed'] = False
            image_path = str(images_dict[image_key]['path'])
            if 'path' in self.ui_element_image_paths[element_name][image_key]:
                if image_path != self.ui_element_image_paths[element_name][image_key]['path']:
                    self.ui_element_image_paths[element_name][image_key]['changed'] = True
            self.ui_element_image_paths[element_name][image_key]['path'] = image_path
            if 'sub_surface_rect' in images_dict[image_key]:
                rect_list = str(images_dict[image_key]['sub_surface_rect']).strip().split(',')
                if len(rect_list) == 4:
                    try:
                        x = int(rect_list[0].strip())
                        y = int(rect_list[1].strip())
                        w = int(rect_list[2].strip())
                        h = int(rect_list[3].strip())
                        rect = pygame.Rect((x, y), (w, h))
                    except ValueError or TypeError:
                        rect = pygame.Rect((0, 0), (10, 10))
                        warnings.warn("Unable to create subsurface rectangle from string: "
                                      "" + images_dict[image_key]['sub_surface_rect'])

                    if 'sub_surface_rect' in self.ui_element_image_paths[element_name][image_key]:
                        if rect != self.ui_element_image_paths[element_name][image_key]['sub_surface_rect']:
                            self.ui_element_image_paths[element_name][image_key]['changed'] = True
                    self.ui_element_image_paths[element_name][image_key]['sub_surface_rect'] = rect

    def load_element_colour_data_from_theme(self, data_type, element_name, theme_dict):
        if element_name not in self.ui_element_colours:
            self.ui_element_colours[element_name] = {}
        colours_dict = theme_dict[element_name][data_type]
        for colour_key in colours_dict:
            self.ui_element_colours[element_name][colour_key] = self.load_colour_or_gradient_from_theme(colours_dict,
                                                                                                        colour_key)

    def load_element_font_data_from_theme(self, data_type, element_name, theme_dict):
        font_dict = theme_dict[element_name][data_type]
        if element_name not in self.ui_element_fonts_info:
            self.ui_element_fonts_info[element_name] = {}
        self.ui_element_fonts_info[element_name]['name'] = font_dict['name']
        try:
            self.ui_element_fonts_info[element_name]['size'] = int(font_dict['size'])
        except ValueError:
            self.ui_element_fonts_info[element_name]['size'] = self.font_dictionary.default_font_size
        if 'bold' in font_dict:
            try:
                self.ui_element_fonts_info[element_name]['bold'] = bool(int(font_dict['bold']))
            except ValueError:
                self.ui_element_fonts_info[element_name]['bold'] = False
        else:
            self.ui_element_fonts_info[element_name]['bold'] = False
        if 'italic' in font_dict:
            try:
                self.ui_element_fonts_info[element_name]['italic'] = bool(int(font_dict['italic']))
            except ValueError:
                self.ui_element_fonts_info[element_name]['italic'] = False
        else:
            self.ui_element_fonts_info[element_name]['italic'] = False
        if 'regular_path' in font_dict:
            self.ui_element_fonts_info[element_name]['regular_path'] = font_dict['regular_path']
        if 'bold_path' in font_dict:
            self.ui_element_fonts_info[element_name]['bold_path'] = font_dict['bold_path']
        if 'italic_path' in font_dict:
            self.ui_element_fonts_info[element_name]['italic_path'] = font_dict['italic_path']
        if 'bold_italic_path' in font_dict:
            bold_italic_path = font_dict['bold_italic_path']
            self.ui_element_fonts_info[element_name]['bold_italic_path'] = bold_italic_path

    def load_colour_defaults_from_theme(self, element_name, theme_dict):
        for data_type in theme_dict[element_name]:
            if data_type == 'colours':
                colours_dict = theme_dict[element_name][data_type]
                for colour_key in colours_dict:
                    self.base_colours[colour_key] = self.load_colour_or_gradient_from_theme(colours_dict, colour_key)

    @staticmethod
    def load_colour_or_gradient_from_theme(theme_colours_dictionary, colour_id):
        loaded_colour_or_gradient = None
        string_data = theme_colours_dictionary[colour_id]
        if ',' in string_data:
            # expecting some type of gradient description in string data
            string_data_list = string_data.split(',')
            gradient_direction = None
            try:
                gradient_direction = int(string_data_list[-1])
            except ValueError:
                warnings.warn("Invalid gradient: " + string_data + " for id:" + colour_id + " in theme file")

            if gradient_direction is not None and len(string_data_list) == 3:
                # two colour gradient
                try:
                    colour_1 = pygame.Color(string_data_list[0])
                    colour_2 = pygame.Color(string_data_list[1])
                    loaded_colour_or_gradient = ColourGradient(gradient_direction, colour_1, colour_2)
                except ValueError:
                    warnings.warn("Invalid gradient: " + string_data + " for id:" + colour_id + " in theme file")
            elif gradient_direction is not None and len(string_data_list) == 4:
                # three colour gradient
                try:
                    colour_1 = pygame.Color(string_data_list[0])
                    colour_2 = pygame.Color(string_data_list[1])
                    colour_3 = pygame.Color(string_data_list[2])
                    loaded_colour_or_gradient = ColourGradient(gradient_direction, colour_1, colour_2, colour_3)
                except ValueError:
                    warnings.warn("Invalid gradient: " + string_data + " for id:" + colour_id + " in theme file")
            else:
                warnings.warn("Invalid gradient: " + string_data + " for id:" + colour_id + " in theme file")
        else:
            # expecting a regular hex colour in string data
            try:
                loaded_colour_or_gradient = pygame.Color(string_data)
            except ValueError:
                warnings.warn("Colour hex code: " + string_data + " for id:" + colour_id + " invalid in theme file")

        if loaded_colour_or_gradient is None:
            # if the colour or gradient data is invalid, return a black default colour
            loaded_colour_or_gradient = pygame.Color("#000000")

        return loaded_colour_or_gradient


class ColourGradient:
    def __init__(self, angle_direction: int, colour_1: pygame.Color,
                 colour_2: pygame.Color, colour_3: Union[pygame.Color, None] = None):
        self.angle_direction = angle_direction
        self.colour_1 = colour_1
        self.colour_2 = colour_2
        self.colour_3 = colour_3

        if self.colour_3 is None:
            pixel_width = 2
            colour_pixels_surf = pygame.Surface((pixel_width, 1), flags=pygame.SRCALPHA, depth=32)
            colour_pixels_surf.fill(self.colour_1, pygame.Rect((0, 0), (1, 1)))
            colour_pixels_surf.fill(self.colour_2, pygame.Rect((1, 0), (1, 1)))
        else:
            pixel_width = 3
            colour_pixels_surf = pygame.Surface((pixel_width, 1), flags=pygame.SRCALPHA, depth=32)
            colour_pixels_surf.fill(self.colour_1, pygame.Rect((0, 0), (1, 1)))
            colour_pixels_surf.fill(self.colour_2, pygame.Rect((1, 0), (1, 1)))
            colour_pixels_surf.fill(self.colour_3, pygame.Rect((2, 0), (1, 1)))

        self.gradient_surface = pygame.transform.rotozoom(colour_pixels_surf, 0, 30)

    def __eq__(self, other):
        if type(other) != ColourGradient:
            return False
        return (self.colour_1 == other.colour_1 and
                self.colour_2 == other.colour_2 and
                self.colour_3 == other.colour_3 and
                self.angle_direction == other.angle_direction)

    def __str__(self):
        result = (str(self.angle_direction) + '_' +
                  str(self.colour_1.r) + '_' + str(self.colour_1.g) + '_' +
                  str(self.colour_1.b) + '_' + str(self.colour_1.a) + '_' +
                  str(self.colour_2.r) + '_' + str(self.colour_2.g) + '_' +
                  str(self.colour_2.b) + '_' + str(self.colour_2.a))
        if self.colour_3 is not None:
            result += ('_' + str(self.colour_3.r) + '_' + str(self.colour_3.g) +
                       '_' + str(self.colour_3.b) + '_' + str(self.colour_3.a))

        return result

    def apply_gradient_to_surface(self, input_surface: pygame.Surface, rect=None):
        """
        Applies this gradient to a specified input surface using blending multiplication.
        As a result this method works best when the input surface is a mostly white, stencil shape type surface.

        :param input_surface:
        :param rect: The rectangle on the surface to apply the gradient to.
        """
        # scale the gradient up to the right size
        input_surface_size = input_surface.get_size()
        inverse_rotated_input = pygame.transform.rotate(input_surface, -self.angle_direction)
        gradient_size = inverse_rotated_input.get_rect().size
        gradient_surf = pygame.Surface(gradient_size, flags=pygame.SRCALPHA, depth=32)

        pygame.transform.scale(self.gradient_surface, gradient_size, gradient_surf)
        gradient_surf = pygame.transform.rotate(gradient_surf, self.angle_direction)

        if rect is not None:
            input_surface.set_clip(rect)
            input_surface.blit(gradient_surf, rect, special_flags=pygame.BLEND_RGBA_MULT)
            input_surface.set_clip(None)
        else:
            gradient_placement_rect = gradient_surf.get_rect()
            gradient_placement_rect.center = (int(input_surface_size[0] / 2), int(input_surface_size[1] / 2))

            input_surface.blit(gradient_surf, gradient_placement_rect, special_flags=pygame.BLEND_RGBA_MULT)


class ShapeCache:
    def __init__(self):
        self.cache_surface_size = (1024, 1024)
        self.cache_surfaces = []
        self.current_surface = pygame.Surface(self.cache_surface_size, flags=pygame.SRCALPHA, depth=32)
        self.current_surface.fill(pygame.Color('#00000000'))
        self.cache_surfaces.append(self.current_surface)

        self.free_space_rectangles = [pygame.Rect((0, 0), self.cache_surface_size)]

        self.cache_long_term_lookup = {}
        self.cache_short_term_lookup = {}

    def add_surface_to_cache(self, surface, string_id):
        self.cache_short_term_lookup[string_id] = surface.copy()

    def update(self):
        if any(self.cache_short_term_lookup):
            string_id, surface = self.cache_short_term_lookup.popitem()
            self.add_surface_to_long_term_cache(surface, string_id)

    def add_surface_to_long_term_cache(self, surface, string_id):
        surface_size = surface.get_size()
        if surface_size > self.cache_surface_size:
            warnings.warn('Unable to cache surfaces larger than ' + str(self.cache_surface_size))
            return None
        else:
            found_rectangle_to_split = None
            found_rectangle_cache = None

            while found_rectangle_cache is None:
                for free_rectangle in self.free_space_rectangles:
                    if found_rectangle_cache is None and (free_rectangle.width >= surface_size[0] and
                                                          free_rectangle.height >= surface_size[1]):
                        # we fits, so we sits
                        found_rectangle_to_split = free_rectangle
                        found_rectangle_cache = pygame.Rect(free_rectangle.topleft, surface_size)
                        self.current_surface.blit(surface, free_rectangle.topleft)
                        self.cache_long_term_lookup[string_id] = self.current_surface.subsurface(found_rectangle_cache)
                        break

                if found_rectangle_to_split is not None and found_rectangle_cache is not None:
                    self.split_rect(found_rectangle_to_split, found_rectangle_cache)
                    # check if our new cached rect overlaps any others
                    rects_to_split = []
                    for rect in self.free_space_rectangles:
                        if rect.colliderect(found_rectangle_cache):
                            rects_to_split.append(rect)

                    for split_rect in rects_to_split:
                        self.split_rect(split_rect, found_rectangle_cache)

                    # clean up rectangles entirely inside other rectangles
                    rects_to_remove = []
                    rectangles_to_check = [rectangle for rectangle in self.free_space_rectangles]
                    for free_rectangle in self.free_space_rectangles:
                        for check_rect in rectangles_to_check:
                            if free_rectangle != check_rect:
                                if check_rect.contains(free_rectangle):
                                    rects_to_remove.append(free_rectangle)
                    self.free_space_rectangles = [rect for rect in self.free_space_rectangles
                                                  if rect not in rects_to_remove]

                elif found_rectangle_to_split is None:
                    # create a new cache surface
                    self.current_surface = pygame.Surface(self.cache_surface_size, flags=pygame.SRCALPHA, depth=32)
                    self.current_surface.fill(pygame.Color('#00000000'))
                    self.cache_surfaces.append(self.current_surface)
                    self.free_space_rectangles = [pygame.Rect((0, 0), self.cache_surface_size)]

            return True

    def split_rect(self, found_rectangle_to_split, dividing_rect):
        self.free_space_rectangles.remove(found_rectangle_to_split)

        # create new rectangles
        if (found_rectangle_to_split.right - dividing_rect.right) > 0:
            rect_1 = pygame.Rect(dividing_rect.right,
                                 found_rectangle_to_split.top,
                                 found_rectangle_to_split.right - dividing_rect.right,
                                 found_rectangle_to_split.height)
            self.free_space_rectangles.append(rect_1)
        if (found_rectangle_to_split.bottom - dividing_rect.bottom) > 0:
            rect_2 = pygame.Rect(found_rectangle_to_split.left,
                                 dividing_rect.bottom,
                                 found_rectangle_to_split.width,
                                 found_rectangle_to_split.bottom - dividing_rect.bottom)
            self.free_space_rectangles.append(rect_2)
        if (dividing_rect.top - found_rectangle_to_split.top) > 0:
            rect_3 = pygame.Rect(found_rectangle_to_split.left,
                                 found_rectangle_to_split.top,
                                 found_rectangle_to_split.width,
                                 dividing_rect.top - found_rectangle_to_split.top)
            self.free_space_rectangles.append(rect_3)
        if (dividing_rect.left - found_rectangle_to_split.left) > 0:
            rect_4 = pygame.Rect(found_rectangle_to_split.left,
                                 found_rectangle_to_split.top,
                                 dividing_rect.left - found_rectangle_to_split.left,
                                 found_rectangle_to_split.height)
            self.free_space_rectangles.append(rect_4)

    def find_surface_in_cache(self, lookup_id):
        # check short term
        if lookup_id in self.cache_short_term_lookup:
            return self.cache_short_term_lookup[lookup_id]
        # check long term
        if lookup_id in self.cache_long_term_lookup:
            return self.cache_long_term_lookup[lookup_id]
        else:
            return None

    @staticmethod
    def build_cache_id(shape, size, shadow_width, border_width, border_colour, bg_colour, corner_radius=None):

        id_string = (shape + '_' + str(size[0]) + '_' + str(size[1]) + '_' +
                     str(shadow_width) + '_' + str(border_width))

        if corner_radius is not None:
            id_string += '_' + str(corner_radius)

        if type(border_colour) == ColourGradient:
            id_string += '_' + str(border_colour)
        else:
            id_string += ('_' + str(border_colour.r) + '_' + str(border_colour.g) +
                          '_' + str(border_colour.b) + '_' + str(border_colour.a))

        if type(bg_colour) == ColourGradient:
            id_string += '_' + str(bg_colour)
        else:
            id_string += ('_' + str(bg_colour.r) + '_' + str(bg_colour.g) +
                          '_' + str(bg_colour.b) + '_' + str(bg_colour.a))

        return id_string



import pygame
import warnings
from typing import Union

from pygame_gui import ui_manager
from pygame_gui.core import ui_container
from pygame_gui.core.ui_element import UIElement
from pygame_gui.core.ui_appearance_theme import ColourGradient


class UILabel(UIElement):
    """
    A label lets us display a single line of text with a single font style. It's a quick to rebuild and simple
    alternative to the text box element.

    :param relative_rect: The rectangle that contains and positions the label relative to it's container.
    :param text: The text to display in the label.
    :param manager: The UIManager that manages this label.
    :param container: The container that this element is within. If set to None will be the root window's container.
    :param parent_element: The element this element 'belongs to' in the theming hierarchy.
    :param object_id: A custom defined ID for fine tuning of theming.
    """
    def __init__(self, relative_rect: pygame.Rect, text: str, manager: ui_manager.UIManager,
                 container: ui_container.UIContainer = None,
                 parent_element: UIElement = None,
                 object_id: Union[str, None] = None):

        new_element_ids, new_object_ids = self.create_valid_ids(parent_element=parent_element,
                                                                object_id=object_id,
                                                                element_id='label')
        super().__init__(relative_rect, manager, container,
                         starting_height=1,
                         layer_thickness=1,
                         object_ids=new_object_ids,
                         element_ids=new_element_ids)
        self.text = text

        # initialise theme params
        self.font = None

        self.bg_colour = None
        self.text_colour = None
        self.text_shadow_colour = None

        self.shadow_enabled = False
        self.shadow_size = 1
        self.shadow_offset = [0, 0]

        self.rebuild_from_changed_theme_data()

    def set_text(self, text: str):
        """
        Changes the string displayed by the label element. Labels do not support HTML styling.

        :param text: the text to set the label to.
        """
        if text != self.text:
            self.text = text
            self.rebuild()

    def rebuild(self):
        """
        Re-render the text to the label's underlying sprite image. This allows us to change what the displayed text is
        or remake it with different theming (if the theming has changed).
        """

        text_size = self.font.size(self.text)
        if text_size[1] > self.rect.height or text_size[0] > self.rect.width:
            width_overlap = self.rect.width - text_size[0]
            height_overlap = self.rect.height - text_size[1]
            warn_text = 'Label Rect is too small for text: ' + self.text + ' - size diff: ' + str((width_overlap,
                                                                                                   height_overlap))
            warnings.warn(warn_text, UserWarning)

        self.image = pygame.Surface(self.rect.size, flags=pygame.SRCALPHA, depth=32)

        if type(self.bg_colour) != ColourGradient and type(self.text_colour) != ColourGradient:
            self.image.fill(self.bg_colour)
            if self.bg_colour.a != 255 or self.shadow_enabled:
                text_render = self.font.render(self.text, True, self.text_colour)
            else:
                text_render = self.font.render(self.text, True, self.text_colour, self.bg_colour)
        else:
            if type(self.bg_colour) != ColourGradient:
                self.image.fill(self.bg_colour)
            else:
                self.image.fill(pygame.Color('#FFFFFFFF'))
                self.bg_colour.apply_gradient_to_surface(self.image)

            if type(self.text_colour) != ColourGradient:
                text_render = self.font.render(self.text, True, self.text_colour)
            else:
                text_render = self.font.render(self.text, True, pygame.Color('#FFFFFFFF'))
                self.text_colour.apply_gradient_to_surface(text_render)

        text_render_rect = text_render.get_rect(centerx=int(self.rect.width / 2),
                                                centery=int(self.rect.height / 2))

        if self.shadow_enabled:
            shadow_text_render = self.font.render(self.text, True, self.text_shadow_colour)

            for y in range(-self.shadow_size, self.shadow_size+1):
                shadow_text_render_rect = pygame.Rect((text_render_rect.x + self.shadow_offset[0],
                                                       text_render_rect.y + self.shadow_offset[1] + y),
                                                      text_render_rect.size)
                self.image.blit(shadow_text_render, shadow_text_render_rect)

            for x in range(-self.shadow_size, self.shadow_size+1):
                shadow_text_render_rect = pygame.Rect((text_render_rect.x + self.shadow_offset[0] + x,
                                                       text_render_rect.y + self.shadow_offset[1]),
                                                      text_render_rect.size)
                self.image.blit(shadow_text_render, shadow_text_render_rect)

            for x_and_y in range(-self.shadow_size, self.shadow_size+1):
                shadow_text_render_rect = pygame.Rect((text_render_rect.x + self.shadow_offset[0] + x_and_y,
                                                       text_render_rect.y + self.shadow_offset[1] + x_and_y),
                                                      text_render_rect.size)
                self.image.blit(shadow_text_render, shadow_text_render_rect)

            for x_and_y in range(-self.shadow_size, self.shadow_size+1):
                shadow_text_render_rect = pygame.Rect((text_render_rect.x + self.shadow_offset[0] - x_and_y,
                                                       text_render_rect.y + self.shadow_offset[1] + x_and_y),
                                                      text_render_rect.size)
                self.image.blit(shadow_text_render, shadow_text_render_rect)

        self.image.blit(text_render, text_render_rect)

    def rebuild_from_changed_theme_data(self):
        any_changed = False

        font = self.ui_theme.get_font(self.object_ids, self.element_ids)
        if font != self.font:
            self.font = font
            any_changed = True

        text_colour = self.ui_theme.get_colour_or_gradient(self.object_ids, self.element_ids, 'normal_text')
        if text_colour != self.text_colour:
            self.text_colour = text_colour
            any_changed = True

        bg_colour = self.ui_theme.get_colour_or_gradient(self.object_ids, self.element_ids, 'dark_bg')
        if bg_colour != self.bg_colour:
            self.bg_colour = bg_colour
            any_changed = True

        text_shadow_colour = self.ui_theme.get_colour(self.object_ids, self.element_ids, 'text_shadow')
        if text_shadow_colour != self.text_shadow_colour:
            self.text_shadow_colour = text_shadow_colour
            any_changed = True

        shadow_enable_param = self.ui_theme.get_misc_data(self.object_ids, self.element_ids, 'text_shadow')
        if shadow_enable_param is not None:
            try:
                shadow_enabled = bool(int(shadow_enable_param))
            except ValueError:
                shadow_enabled = False
            if shadow_enabled != self.shadow_enabled:
                self.shadow_enabled = shadow_enabled
                any_changed = True

        shadow_size_param = self.ui_theme.get_misc_data(self.object_ids, self.element_ids, 'text_shadow_size')
        if shadow_size_param is not None:
            try:
                shadow_size = int(shadow_size_param)
            except ValueError:
                shadow_size = 1
            if shadow_size != self.shadow_size:
                self.shadow_size = shadow_size
                any_changed = True

        shadow_offset_param = self.ui_theme.get_misc_data(self.object_ids, self.element_ids, 'text_shadow_offset')
        if shadow_offset_param is not None:
            offset_string_list = shadow_offset_param.split(',')
            if len(offset_string_list) == 2:
                try:
                    shadow_offset = [int(offset_string_list[0]), int(offset_string_list[1])]
                except ValueError:
                    shadow_offset = [1, 1]
                if shadow_offset != self.shadow_offset:
                    self.shadow_offset = shadow_offset
                    any_changed = True

        if any_changed:
            self.rebuild()
